from __future__ import (
    annotations,
)

import base64
from collections.abc import (
    Callable,
    KeysView,
)
import functools
import hashlib
import logging
import time
from typing import (
    NamedTuple,
    cast,
)

import base58
import trio

from nanolib.abc import (
    IHost,
    INetStream,
    IPubsub,
    IPubsubRouter,
    ISubscriptionAPI,
)
from nanolib.crypto.keys import (
    PrivateKey,
)
from nanolib.custom_types import (
    AsyncValidatorFn,
    SyncValidatorFn,
    TProtocol,
    ValidatorFn,
)
from nanolib.exceptions import (
    ParseError,
    ValidationError,
)
from nanolib.io.exceptions import (
    IncompleteReadError,
)
from nanolib.network.exceptions import (
    SwarmException,
)
from nanolib.network.stream.exceptions import (
    StreamClosed,
    StreamEOF,
    StreamError,
)
from nanolib.peer.id import (
    ID,
)
from nanolib.peer.peerdata import (
    PeerDataError,
)
from nanolib.peer.peerstore import env_to_send_in_RPC
from nanolib.pubsub.utils import maybe_consume_signed_record
from nanolib.tools.async_service import (
    Service,
)
from nanolib.tools.timed_cache.last_seen_cache import (
    LastSeenCache,
)
from nanolib.utils import (
    encode_varint_prefixed,
    read_varint_prefixed_bytes,
)
from nanolib.utils.varint import encode_uvarint

from .pb import (
    rpc_pb2,
)
from .pubsub_notifee import (
    PubsubNotifee,
)
from .subscription import (
    TrioSubscriptionAPI,
)
from .validators import (
    PUBSUB_SIGNING_PREFIX,
    signature_validator,
)

# Ref: https://github.com/libp2p/go-libp2p-pubsub/blob/40e1c94708658b155f30cf99e4574f384756d83c/topic.go#L97  # noqa: E501
SUBSCRIPTION_CHANNEL_SIZE = 32

logger = logging.getLogger("nanolib.pubsub")


def get_peer_and_seqno_msg_id(msg: rpc_pb2.Message) -> bytes:
    # NOTE: `string(from, seqno)` in Go
    return msg.seqno + msg.from_id


def get_content_addressed_msg_id(msg: rpc_pb2.Message) -> bytes:
    return base64.b64encode(hashlib.sha256(msg.data).digest())


class TopicValidator(NamedTuple):
    validator: ValidatorFn
    is_async: bool


MAX_CONCURRENT_VALIDATORS = 10


class Pubsub(Service, IPubsub):
    host: IHost

    router: IPubsubRouter

    peer_receive_channel: trio.MemoryReceiveChannel[ID]
    dead_peer_receive_channel: trio.MemoryReceiveChannel[ID]
    _validator_semaphore: trio.Semaphore

    seen_messages: LastSeenCache

    subscribed_topics_send: dict[str, trio.MemorySendChannel[rpc_pb2.Message]]
    subscribed_topics_receive: dict[str, TrioSubscriptionAPI]

    peer_topics: dict[str, set[ID]]
    peers: dict[ID, INetStream]

    topic_validators: dict[str, TopicValidator]

    counter: int  # uint64

    # Indicate if we should enforce signature verification
    strict_signing: bool
    sign_key: PrivateKey | None

    # Set of blacklisted peer IDs
    blacklisted_peers: set[ID]

    event_handle_peer_queue_started: trio.Event
    event_handle_dead_peer_queue_started: trio.Event

    def __init__(
        self,
        host: IHost,
        router: IPubsubRouter,
        cache_size: int | None = None,
        seen_ttl: int = 120,
        sweep_interval: int = 60,
        strict_signing: bool = True,
        msg_id_constructor: Callable[
            [rpc_pb2.Message], bytes
        ] = get_peer_and_seqno_msg_id,
        max_concurrent_validator_count: int = MAX_CONCURRENT_VALIDATORS,
    ) -> None:
        """
        Construct a new Pubsub object, which is responsible for handling all
        Pubsub-related messages and relaying messages as appropriate to the
        Pubsub router (which is responsible for choosing who to send messages
        to).

        Since the logic for choosing peers to send pubsub messages to is
        in the router, the same Pubsub impl can back floodsub,
        gossipsub, etc.
        """
        self.host = host
        self.router = router

        self._msg_id_constructor = msg_id_constructor

        # Attach this new Pubsub object to the router
        self.router.attach(self)

        peer_send, peer_receive = trio.open_memory_channel[ID](0)
        dead_peer_send, dead_peer_receive = trio.open_memory_channel[ID](0)
        # Only keep the receive channels in `Pubsub`.
        # Therefore, we can only close from the receive side.
        self.peer_receive_channel = peer_receive
        self.dead_peer_receive_channel = dead_peer_receive
        self._validator_semaphore = trio.Semaphore(max_concurrent_validator_count)
        # Register a notifee
        self.host.get_network().register_notifee(
            PubsubNotifee(peer_send, dead_peer_send)
        )

        # Register stream handlers for each pubsub router protocol to handle
        # the pubsub streams opened on those protocols
        for protocol in router.get_protocols():
            self.host.set_stream_handler(protocol, self.stream_handler)

        # keeps track of seen messages as LRU cache
        if cache_size is None:
            self.cache_size = 128
        else:
            self.cache_size = cache_size

        self.strict_signing = strict_signing
        if strict_signing:
            self.sign_key = self.host.get_private_key()
        else:
            self.sign_key = None

        self.seen_messages = LastSeenCache(seen_ttl, sweep_interval)

        # Map of topics we are subscribed to blocking queues
        # for when the given topic receives a message
        self.subscribed_topics_send = {}
        self.subscribed_topics_receive = {}

        # Map of topic to peers to keep track of what peers are subscribed to
        self.peer_topics = {}

        # Create peers map, which maps peer_id (as string) to stream (to a given peer)
        self.peers = {}

        # Map of topic to topic validator
        self.topic_validators = {}

        self.counter = int(time.time())

        # Set of blacklisted peer IDs
        self.blacklisted_peers = set()

        self.event_handle_peer_queue_started = trio.Event()
        self.event_handle_dead_peer_queue_started = trio.Event()

    async def run(self) -> None:
        self.manager.run_daemon_task(self.handle_peer_queue)
        self.manager.run_daemon_task(self.handle_dead_peer_queue)
        await self.manager.wait_finished()

    @property
    def my_id(self) -> ID:
        return self.host.get_id()

    @property
    def protocols(self) -> tuple[TProtocol, ...]:
        return tuple(self.router.get_protocols())

    @property
    def topic_ids(self) -> KeysView[str]:
        return self.subscribed_topics_receive.keys()

    def get_hello_packet(self) -> rpc_pb2.RPC:
        """
        Generate subscription message with all topics we are subscribed to
        only send hello packet if we have subscribed topics.
        """
        packet = rpc_pb2.RPC()
        for topic_id in self.topic_ids:
            packet.subscriptions.extend(
                [rpc_pb2.RPC.SubOpts(subscribe=True, topicid=topic_id)]
            )
        # Add the sender's signedRecord in the RPC message
        envelope_bytes, _ = env_to_send_in_RPC(self.host)
        packet.senderRecord = envelope_bytes

        return packet

    async def continuously_read_stream(self, stream: INetStream) -> None:
        """
        Read from input stream in an infinite loop. Process messages from other
        nodes.

        :param stream: stream to continously read from
        """
        peer_id = stream.muxed_conn.peer_id

        try:
            while self.manager.is_running:
                incoming: bytes = await read_varint_prefixed_bytes(stream)
                rpc_incoming: rpc_pb2.RPC = rpc_pb2.RPC()
                rpc_incoming.ParseFromString(incoming)

                # Process the sender's signed-record if sent
                if not maybe_consume_signed_record(rpc_incoming, self.host, peer_id):
                    logger.error(
                        "Received an invalid-signed-record, ignoring the incoming msg"
                    )
                    continue

                if rpc_incoming.publish:
                    # deal with RPC.publish
                    for msg in rpc_incoming.publish:
                        if not self._is_subscribed_to_msg(msg):
                            continue
                        logger.debug(
                            "received `publish` message %s from peer %s", msg, peer_id
                        )
                        # Only schedule task if service is still running
                        if self.manager.is_running:
                            self.manager.run_task(self.push_msg, peer_id, msg)

                if rpc_incoming.subscriptions:
                    # deal with RPC.subscriptions
                    # We don't need to relay the subscription to our
                    # peers because a given node only needs its peers
                    # to know that it is subscribed to the topic (doesn't
                    # need everyone to know)
                    for message in rpc_incoming.subscriptions:
                        logger.debug(
                            "received `subscriptions` message %s from peer %s",
                            message,
                            peer_id,
                        )
                        self.handle_subscription(peer_id, message)

                # NOTE: Check if `rpc_incoming.control` is set through `HasField`.
                #   This is necessary because `control` is an optional field in pb2.
                #   Ref: https://developers.google.com/protocol-buffers/docs/reference/python-generated#singular-fields-proto2  # noqa: E501
                if rpc_incoming.HasField("control"):
                    # Pass rpc to router so router could perform custom logic
                    logger.debug(
                        "received `control` message %s from peer %s",
                        rpc_incoming.control,
                        peer_id,
                    )
                    await self.router.handle_rpc(rpc_incoming, peer_id)
        except StreamEOF:
            logger.debug(
                f"Stream closed for peer {peer_id}, exiting read loop cleanly."
            )

    def set_topic_validator(
        self, topic: str, validator: ValidatorFn, is_async_validator: bool
    ) -> None:
        """
        Register a validator under the given topic. One topic can only have one
        validtor.

        :param topic: the topic to register validator under
        :param validator: the validator used to validate messages published to the topic
        :param is_async_validator: indicate if the validator is an asynchronous validator
        """  # noqa: E501
        self.topic_validators[topic] = TopicValidator(validator, is_async_validator)

    def remove_topic_validator(self, topic: str) -> None:
        """
        Remove the validator from the given topic.

        :param topic: the topic to remove validator from
        """
        self.topic_validators.pop(topic, None)

    def get_msg_validators(self, msg: rpc_pb2.Message) -> tuple[TopicValidator, ...]:
        """
        Get all validators corresponding to the topics in the message.

        :param msg: the message published to the topic
        """
        return tuple(
            self.topic_validators[topic]
            for topic in msg.topicIDs
            if topic in self.topic_validators
        )

    def add_to_blacklist(self, peer_id: ID) -> None:
        """
        Add a peer to the blacklist.
        When a peer is blacklisted:
        - Any existing connection to that peer is immediately closed and removed
        - The peer is removed from all topic subscription mappings
        - Future connection attempts from this peer will be rejected
        - Messages forwarded by or originating from this peer will be dropped
        - The peer will not be able to participate in pubsub communication

        :param peer_id: the peer ID to blacklist
        """
        self.blacklisted_peers.add(peer_id)
        logger.debug("Added peer %s to blacklist", peer_id)
        self.manager.run_task(self._teardown_if_connected, peer_id)

    async def _teardown_if_connected(self, peer_id: ID) -> None:
        """Close their stream and remove them if connected"""
        stream = self.peers.get(peer_id)
        if stream is not None:
            try:
                await stream.reset()
            except Exception:
                pass
            del self.peers[peer_id]
        # Also remove from any subscription maps:
        for _topic, peerset in self.peer_topics.items():
            if peer_id in peerset:
                peerset.discard(peer_id)

    def remove_from_blacklist(self, peer_id: ID) -> None:
        """
        Remove a peer from the blacklist.
        Once removed from the blacklist:
        - The peer can establish new connections to this node
        - Messages from this peer will be processed normally
        - The peer can participate in topic subscriptions and message forwarding

        :param peer_id: the peer ID to remove from blacklist
        """
        self.blacklisted_peers.discard(peer_id)
        logger.debug("Removed peer %s from blacklist", peer_id)

    def is_peer_blacklisted(self, peer_id: ID) -> bool:
        """
        Check if a peer is blacklisted.

        :param peer_id: the peer ID to check
        :return: True if peer is blacklisted, False otherwise
        """
        return peer_id in self.blacklisted_peers

    def clear_blacklist(self) -> None:
        """
        Clear all peers from the blacklist.
        This removes all blacklist restrictions, allowing previously blacklisted
        peers to:
        - Establish new connections
        - Send and forward messages
        - Participate in topic subscriptions

        """
        self.blacklisted_peers.clear()
        logger.debug("Cleared all peers from blacklist")

    def get_blacklisted_peers(self) -> set[ID]:
        """
        Get a copy of the current blacklisted peers.
        Returns a snapshot of all currently blacklisted peer IDs. These peers
        are completely isolated from pubsub communication - their connections
        are rejected and their messages are dropped.

        :return: a set containing all blacklisted peer IDs
        """
        return self.blacklisted_peers.copy()

    async def stream_handler(self, stream: INetStream) -> None:
        """
        Stream handler for pubsub. Gets invoked whenever a new stream is
        created on one of the supported pubsub protocols.

        :param stream: newly created stream
        """
        peer_id = stream.muxed_conn.peer_id

        try:
            await self.continuously_read_stream(stream)
        except (StreamError, ParseError, IncompleteReadError) as error:
            logger.debug(
                "fail to read from peer %s, error=%s,"
                "closing the stream and remove the peer from record",
                peer_id,
                error,
            )
            await stream.reset()
            self._handle_dead_peer(peer_id)

    async def wait_until_ready(self) -> None:
        await self.event_handle_peer_queue_started.wait()
        await self.event_handle_dead_peer_queue_started.wait()

    async def _handle_new_peer(self, peer_id: ID) -> None:
        if self.is_peer_blacklisted(peer_id):
            logger.debug("Rejecting blacklisted peer %s", peer_id)
            return

        try:
            stream: INetStream = await self.host.new_stream(peer_id, self.protocols)
        except SwarmException as error:
            logger.debug("fail to add new peer %s, error %s", peer_id, error)
            return

        # Send hello packet
        hello = self.get_hello_packet()
        try:
            await stream.write(encode_varint_prefixed(hello.SerializeToString()))
        except StreamClosed:
            logger.debug("Fail to add new peer %s: stream closed", peer_id)
            return
        try:
            self.router.add_peer(peer_id, stream.get_protocol())
        except Exception as error:
            logger.debug("fail to add new peer %s, error %s", peer_id, error)
            return

        self.peers[peer_id] = stream

        logger.debug("added new peer %s", peer_id)

    async def _handle_new_peer_safe(self, peer_id: ID) -> None:
        """
        Safely handle new peer with exception handling.
        This wrapper ensures that any exceptions during peer negotiation
        don't crash the entire pubsub service.
        """
        try:
            await self._handle_new_peer(peer_id)
        except Exception as error:
            logger.info(f"Protocol negotiation failed for peer {peer_id}: {error}")

    def _handle_dead_peer(self, peer_id: ID) -> None:
        if peer_id not in self.peers:
            return
        del self.peers[peer_id]

        for topic in self.peer_topics:
            if peer_id in self.peer_topics[topic]:
                self.peer_topics[topic].discard(peer_id)

        self.router.remove_peer(peer_id)

        logger.debug("removed dead peer %s", peer_id)

    async def handle_peer_queue(self) -> None:
        """
        Continuously read from peer queue and each time a new peer is found,
        open a stream to the peer using a supported pubsub protocol pubsub
        protocols we support.
        """
        async with self.peer_receive_channel:
            self.event_handle_peer_queue_started.set()
            async for peer_id in self.peer_receive_channel:
                # Add Peer - wrap in exception handler to prevent service crash
                self.manager.run_task(self._handle_new_peer_safe, peer_id)

    async def handle_dead_peer_queue(self) -> None:
        """
        Continuously read from dead peer channel and close the stream
        between that peer and remove peer info from pubsub and pubsub router.
        """
        async with self.dead_peer_receive_channel:
            self.event_handle_dead_peer_queue_started.set()
            async for peer_id in self.dead_peer_receive_channel:
                # Remove Peer
                self._handle_dead_peer(peer_id)

    def handle_subscription(
        self, origin_id: ID, sub_message: rpc_pb2.RPC.SubOpts
    ) -> None:
        """
        Handle an incoming subscription message from a peer. Update internal
        mapping to mark the peer as subscribed or unsubscribed to topics as
        defined in the subscription message.

        :param origin_id: id of the peer who subscribe to the message
        :param sub_message: RPC.SubOpts
        """
        if sub_message.subscribe:
            if sub_message.topicid not in self.peer_topics:
                self.peer_topics[sub_message.topicid] = {origin_id}
            elif origin_id not in self.peer_topics[sub_message.topicid]:
                # Add peer to topic
                self.peer_topics[sub_message.topicid].add(origin_id)
        else:
            if sub_message.topicid in self.peer_topics:
                if origin_id in self.peer_topics[sub_message.topicid]:
                    self.peer_topics[sub_message.topicid].discard(origin_id)

    def notify_subscriptions(self, publish_message: rpc_pb2.Message) -> None:
        """
        Put incoming message from a peer onto my blocking queue.

        :param publish_message: RPC.Message format
        """
        # Check if this message has any topics that we are subscribed to
        for topic in publish_message.topicIDs:
            if topic in self.topic_ids:
                # we are subscribed to a topic this message was sent for,
                # so add message to the subscription output queue
                # for each topic
                try:
                    self.subscribed_topics_send[topic].send_nowait(publish_message)
                except trio.WouldBlock:
                    # Channel is full, ignore this message.
                    logger.warning(
                        "fail to deliver message to subscription for topic %s", topic
                    )

    async def subscribe(self, topic_id: str) -> ISubscriptionAPI:
        """
        Subscribe ourself to a topic.

        :param topic_id: topic_id to subscribe to
        """
        logger.debug("subscribing to topic %s", topic_id)

        # Already subscribed
        if topic_id in self.topic_ids:
            return self.subscribed_topics_receive[topic_id]

        send_channel, receive_channel = trio.open_memory_channel[rpc_pb2.Message](
            SUBSCRIPTION_CHANNEL_SIZE
        )

        subscription = TrioSubscriptionAPI(
            receive_channel,
            unsubscribe_fn=functools.partial(self.unsubscribe, topic_id),
        )
        self.subscribed_topics_send[topic_id] = send_channel
        self.subscribed_topics_receive[topic_id] = subscription

        # Create subscribe message
        packet: rpc_pb2.RPC = rpc_pb2.RPC()
        packet.subscriptions.extend(
            [rpc_pb2.RPC.SubOpts(subscribe=True, topicid=topic_id)]
        )

        # Add the senderRecord of the peer in the RPC msg
        envelope_bytes, _ = env_to_send_in_RPC(self.host)
        packet.senderRecord = envelope_bytes
        # Send out subscribe message to all peers
        await self.message_all_peers(packet.SerializeToString())

        # Tell router we are joining this topic
        await self.router.join(topic_id)

        # Return the subscription for messages on this topic
        return subscription

    async def unsubscribe(self, topic_id: str) -> None:
        """
        Unsubscribe ourself from a topic.

        :param topic_id: topic_id to unsubscribe from
        """
        logger.debug("unsubscribing from topic %s", topic_id)

        # Return if we already unsubscribed from the topic
        if topic_id not in self.topic_ids:
            return
        # Remove topic_id from the maps before yielding
        send_channel = self.subscribed_topics_send[topic_id]
        del self.subscribed_topics_send[topic_id]
        del self.subscribed_topics_receive[topic_id]
        # Only close the send side
        await send_channel.aclose()

        # Create unsubscribe message
        packet: rpc_pb2.RPC = rpc_pb2.RPC()
        packet.subscriptions.extend(
            [rpc_pb2.RPC.SubOpts(subscribe=False, topicid=topic_id)]
        )
        # Add the senderRecord of the peer in the RPC msg
        envelope_bytes, _ = env_to_send_in_RPC(self.host)
        packet.senderRecord = envelope_bytes

        # Send out unsubscribe message to all peers
        await self.message_all_peers(packet.SerializeToString())

        # Tell router we are leaving this topic
        await self.router.leave(topic_id)

    async def message_all_peers(self, raw_msg: bytes) -> None:
        """
        Broadcast a message to peers.

        :param raw_msg: raw contents of the message to broadcast
        """
        # Broadcast message
        for stream in self.peers.values():
            # Write message to stream
            try:
                await stream.write(encode_varint_prefixed(raw_msg))
            except StreamClosed:
                peer_id = stream.muxed_conn.peer_id
                logger.debug("Fail to message peer %s: stream closed", peer_id)
                self._handle_dead_peer(peer_id)

    async def publish(self, topic_id: str | list[str], data: bytes) -> None:
        """
        Publish data to a topic or multiple topics.

        :param topic_id: topic (str) or topics (list[str]) to publish the data to
        :param data: data which we are publishing
        """
        # Handle both single topic (str) and multiple topics (list[str])
        if isinstance(topic_id, str):
            topic_ids = [topic_id]
        else:
            topic_ids = topic_id

        msg = rpc_pb2.Message(
            data=data,
            topicIDs=topic_ids,
            # Origin is ourself.
            from_id=self.my_id.to_bytes(),
            seqno=self._next_seqno(),
        )

        if self.strict_signing:
            priv_key = self.sign_key
            if priv_key is None:
                raise PeerDataError("private key not found")

            signature = priv_key.sign(
                PUBSUB_SIGNING_PREFIX.encode() + msg.SerializeToString()
            )
            msg.key = self.host.get_public_key().serialize()
            msg.signature = signature

        await self.push_msg(self.my_id, msg)

        logger.debug("successfully published message %s", msg)

    async def validate_msg(
        self,
        msg_forwarder: ID,
        msg: rpc_pb2.Message,
    ) -> None:
        """
        Validate the received message.

        :param msg_forwarder: the peer who forward us the message.
        :param msg: the message.
        """
        sync_topic_validators: list[SyncValidatorFn] = []
        async_topic_validators: list[AsyncValidatorFn] = []
        for topic_validator in self.get_msg_validators(msg):
            if topic_validator.is_async:
                async_topic_validators.append(
                    cast(AsyncValidatorFn, topic_validator.validator)
                )
            else:
                sync_topic_validators.append(
                    cast(SyncValidatorFn, topic_validator.validator)
                )

        for validator in sync_topic_validators:
            if not validator(msg_forwarder, msg):
                raise ValidationError(f"Validation failed for msg={msg}")

        if len(async_topic_validators) > 0:
            # Appends to lists are thread safe in CPython
            results: list[bool] = []

            async with trio.open_nursery() as nursery:
                for async_validator in async_topic_validators:
                    nursery.start_soon(
                        self._run_async_validator,
                        async_validator,
                        msg_forwarder,
                        msg,
                        results,
                    )

            if not all(results):
                raise ValidationError(f"Validation failed for msg={msg}")

    async def _run_async_validator(
        self,
        func: AsyncValidatorFn,
        msg_forwarder: ID,
        msg: rpc_pb2.Message,
        results: list[bool],
    ) -> None:
        async with self._validator_semaphore:
            result = await func(msg_forwarder, msg)
            results.append(result)

    async def push_msg(self, msg_forwarder: ID, msg: rpc_pb2.Message) -> None:
        """
        Push a pubsub message to others.

        :param msg_forwarder: the peer who forward us the message.
        :param msg: the message we are going to push out.
        """
        logger.debug("attempting to publish message %s", msg)

        # Check if the message forwarder (source) is in the blacklist. If yes, reject.
        if self.is_peer_blacklisted(msg_forwarder):
            logger.debug(
                "Rejecting message from blacklisted source peer %s", msg_forwarder
            )
            return

        # Check if the message originator (from) is in the blacklist. If yes, reject.
        msg_from_peer = ID(msg.from_id)
        if self.is_peer_blacklisted(msg_from_peer):
            logger.debug(
                "Rejecting message from blacklisted originator peer %s", msg_from_peer
            )
            return

        # If the message is processed before, return(i.e., don't further process the message)  # noqa: E501
        if self._is_msg_seen(msg):
            return

        try:
            scorer = getattr(self.router, "scorer", None)
            if scorer is not None:
                if not scorer.allow_publish(msg_forwarder, list(msg.topicIDs)):
                    logger.debug(
                        "Rejecting message from %s by publish score gate", msg_forwarder
                    )
                    return
        except Exception:
            # Router may not support scoring; ignore gracefully
            pass

        # Check if signing is required and if so validate the signature
        if self.strict_signing:
            # Validate the signature of the message
            if not signature_validator(msg):
                logger.debug("Signature validation failed for msg: %s", msg)
                return

        # Validate the message with registered topic validators.
        # If the validation failed, return(i.e., don't further process the message).
        try:
            await self.validate_msg(msg_forwarder, msg)
        except ValidationError:
            # Scoring: count invalid messages
            try:
                scorer = getattr(self.router, "scorer", None)
                if scorer is not None:
                    for topic in msg.topicIDs:
                        scorer.on_invalid_message(msg_forwarder, topic)
            except Exception:
                pass
            logger.debug(
                "Topic validation failed: sender %s sent data %s under topic IDs: %s %s:%s",  # noqa: E501
                msg_forwarder,
                msg.data.hex(),
                msg.topicIDs,
                base58.b58encode(msg.from_id).decode(),
                msg.seqno.hex(),
            )
            return

        self._mark_msg_seen(msg)

        # Scoring: first delivery for this sender per topic
        try:
            scorer = getattr(self.router, "scorer", None)
            if scorer is not None:
                for topic in msg.topicIDs:
                    scorer.on_first_delivery(msg_forwarder, topic)
        except Exception:
            pass

        # reject messages claiming to be from ourselves but not locally published
        self_id = self.host.get_id()
        if (
            base58.b58encode(msg.from_id).decode() == self_id
            and msg_forwarder != self_id
        ):
            logger.debug(
                "dropping message claiming to be from self but forwarded from %s",
                msg_forwarder,
            )
            return

        self.notify_subscriptions(msg)
        await self.router.publish(msg_forwarder, msg)

    def _next_seqno(self) -> bytes:
        """Make the next message sequence id."""
        self.counter += 1
        return self.counter.to_bytes(8, "big")

    def _is_msg_seen(self, msg: rpc_pb2.Message) -> bool:
        msg_id = self._msg_id_constructor(msg)
        return self.seen_messages.has(msg_id)

    def _mark_msg_seen(self, msg: rpc_pb2.Message) -> None:
        msg_id = self._msg_id_constructor(msg)
        self.seen_messages.add(msg_id)

    def _is_subscribed_to_msg(self, msg: rpc_pb2.Message) -> bool:
        return any(topic in self.topic_ids for topic in msg.topicIDs)

    def get_message_id(self, msg: rpc_pb2.Message) -> bytes:
        """
        Get the message ID for a given message using the configured
        message ID constructor.

        This method provides a public interface for external components (like routers)
        to access message ID construction functionality.

        :param msg: the message to get the ID for
        :return: the message ID as bytes
        """
        return self._msg_id_constructor(msg)

    async def write_msg(self, stream: INetStream, rpc_msg: rpc_pb2.RPC) -> bool:
        """
        Write an RPC message to a stream with proper error handling.

        Implements WriteMsg similar to go-msgio which is used in go-libp2p
        Ref: https://github.com/libp2p/go-msgio/blob/master/protoio/uvarint_writer.go#L56


        :param stream: stream to write the message to
        :param rpc_msg: RPC message to write
        :return: True if successful, False if stream was closed
        """
        try:
            # Calculate message size first
            msg_bytes = rpc_msg.SerializeToString()
            msg_size = len(msg_bytes)

            # Calculate varint size and allocate exact buffer size needed

            varint_bytes = encode_uvarint(msg_size)
            varint_size = len(varint_bytes)

            # Allocate buffer with exact size (like Go's pool.Get())
            buf = bytearray(varint_size + msg_size)

            # Write varint length prefix to buffer (like Go's binary.PutUvarint())
            buf[:varint_size] = varint_bytes

            # Write serialized message after varint (like Go's rpc.MarshalTo())
            buf[varint_size:] = msg_bytes

            # Single write operation (like Go's s.Write(buf))
            await stream.write(bytes(buf))
            return True
        except StreamClosed:
            peer_id = stream.muxed_conn.peer_id
            logger.debug("Fail to write message to %s: stream closed", peer_id)
            self._handle_dead_peer(peer_id)
            return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: libp2p/peer/pb/envelope.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from nanolib.peer.pb import crypto_pb2 as libp2p_dot_peer_dot_pb_dot_crypto__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dlibp2p/peer/pb/envelope.proto\x12\x15libp2p.peer.pb.record\x1a\x1blibp2p/peer/pb/crypto.proto\"z\n\x08\x45nvelope\x12\x34\n\npublic_key\x18\x01 \x01(\x0b\x32 .libp2p.peer.pb.crypto.PublicKey\x12\x14\n\x0cpayload_type\x18\x02 \x01(\x0c\x12\x0f\n\x07payload\x18\x03 \x01(\x0c\x12\x11\n\tsignature\x18\x05 \x01(\x0c\x42,Z*github.com/libp2p/go-libp2p/core/record/pbb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.peer.pb.envelope_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z*github.com/libp2p/go-libp2p/core/record/pb'
  _globals['_ENVELOPE']._serialized_start=85
  _globals['_ENVELOPE']._serialized_end=207
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass


@dataclass
class RetryConfig:
    """
    Configuration for retry logic with exponential backoff.

    This configuration controls how connection attempts are retried when they fail.
    The retry mechanism uses exponential backoff with jitter to prevent thundering
    herd problems in distributed systems.

    Attributes:
        max_retries: Maximum number of retry attempts before giving up.
                     Default: 3 attempts
        initial_delay: Initial delay in seconds before the first retry.
                      Default: 0.1 seconds (100ms)
        max_delay: Maximum delay cap in seconds to prevent excessive wait times.
                  Default: 30.0 seconds
        backoff_multiplier: Multiplier for exponential backoff (each retry multiplies
                           the delay by this factor). Default: 2.0 (doubles each time)
        jitter_factor: Random jitter factor (0.0-1.0) to add randomness to delays
                      and prevent synchronized retries. Default: 0.1 (10% jitter)

    """

    max_retries: int = 3
    initial_delay: float = 0.1
    max_delay: float = 30.0
    backoff_multiplier: float = 2.0
    jitter_factor: float = 0.1


@dataclass
class ConnectionConfig:
    """
    Configuration for multi-connection support.

    This configuration controls how multiple connections per peer are managed,
    including connection limits, timeouts, and load balancing strategies.

    Attributes:
        max_connections_per_peer: Maximum number of connections allowed to a single
                                 peer. Default: 3 connections
        connection_timeout: Timeout in seconds for establishing new connections.
                           Default: 30.0 seconds
        load_balancing_strategy: Strategy for distributing streams across connections.
                                Options: "round_robin" (default) or "least_loaded"

    """

    max_connections_per_peer: int = 3
    connection_timeout: float = 30.0
    load_balancing_strategy: str = "round_robin"  # or "least_loaded"

    def __post_init__(self) -> None:
        """Validate configuration after initialization."""
        if not (
            self.load_balancing_strategy == "round_robin"
            or self.load_balancing_strategy == "least_loaded"
        ):
            raise ValueError(
                "Load balancing strategy can only be 'round_robin' or 'least_loaded'"
            )

        if self.max_connections_per_peer < 1:
            raise ValueError("Max connection per peer should be atleast 1")

        if self.connection_timeout < 0:
            raise ValueError("Connection timeout should be positive")



"""
Safe serialization utilities for persistent peerstore.

This module provides safe serialization/deserialization functions using Protocol Buffers
instead of pickle to avoid security vulnerabilities.
"""

from collections.abc import Sequence
import logging
from typing import Any

from multiaddr import Multiaddr

from nanolib.crypto.keys import KeyPair
from nanolib.peer.envelope import Envelope
from nanolib.peer.pb.crypto_pb2 import (
    KeyType as PBKeyType,
    PrivateKey as PBPrivateKey,
    PublicKey as PBPublicKey,
)
from nanolib.peer.peerstore import PeerRecordState

from .pb import (
    PeerAddresses,
    PeerEnvelope,
    PeerKeys,
    PeerLatency,
    PeerMetadata,
    PeerProtocols,
    PeerRecordState as PBPeerRecordState,
)

logger = logging.getLogger(__name__)


class SerializationError(Exception):
    """Raised when serialization or deserialization fails."""


def serialize_addresses(addresses: Sequence[Multiaddr]) -> bytes:
    """
    Serialize a sequence of multiaddresses to bytes.

    :param addresses: Sequence of Multiaddr objects to serialize
    :return: Serialized bytes
    :raises SerializationError: If serialization fails
    """
    try:
        pb_addresses = PeerAddresses()
        pb_addresses.addresses.extend([addr.to_bytes() for addr in addresses])
        return pb_addresses.SerializeToString()
    except Exception as e:
        raise SerializationError(f"Failed to serialize addresses: {e}") from e


def deserialize_addresses(data: bytes) -> list[Multiaddr]:
    """
    Deserialize addresses from bytes.

    :param data: Serialized address data
    :return: List of Multiaddr objects
    :raises SerializationError: If deserialization fails
    """
    try:
        pb_addresses = PeerAddresses()
        pb_addresses.ParseFromString(data)
        return [Multiaddr(addr_bytes) for addr_bytes in pb_addresses.addresses]
    except Exception as e:
        raise SerializationError(f"Failed to deserialize addresses: {e}") from e


def serialize_protocols(protocols: Sequence[str]) -> bytes:
    """
    Serialize a sequence of protocol strings to bytes.

    :param protocols: Sequence of protocol strings to serialize
    :return: Serialized bytes
    :raises SerializationError: If serialization fails
    """
    try:
        pb_protocols = PeerProtocols()
        pb_protocols.protocols.extend(protocols)
        return pb_protocols.SerializeToString()
    except Exception as e:
        raise SerializationError(f"Failed to serialize protocols: {e}") from e


def deserialize_protocols(data: bytes) -> list[str]:
    """
    Deserialize protocols from bytes.

    :param data: Serialized protocol data
    :return: List of protocol strings
    :raises SerializationError: If deserialization fails
    """
    try:
        pb_protocols = PeerProtocols()
        pb_protocols.ParseFromString(data)
        return list(pb_protocols.protocols)
    except Exception as e:
        raise SerializationError(f"Failed to deserialize protocols: {e}") from e


def serialize_metadata(metadata: dict[str, Any]) -> bytes:
    """
    Serialize metadata dictionary to bytes.

    :param metadata: Dictionary of metadata to serialize
    :return: Serialized bytes
    :raises SerializationError: If serialization fails
    """
    try:
        pb_metadata = PeerMetadata()
        for key, value in metadata.items():
            # Convert value to bytes if it's not already
            if isinstance(value, bytes):
                pb_metadata.metadata[key] = value
            elif isinstance(value, str):
                pb_metadata.metadata[key] = value.encode("utf-8")
            elif isinstance(value, (int, float)):
                pb_metadata.metadata[key] = str(value).encode("utf-8")
            else:
                # For other types, convert to string representation
                pb_metadata.metadata[key] = str(value).encode("utf-8")
        return pb_metadata.SerializeToString()
    except Exception as e:
        raise SerializationError(f"Failed to serialize metadata: {e}") from e


def deserialize_metadata(data: bytes) -> dict[str, bytes]:
    """
    Deserialize metadata from bytes.

    :param data: Serialized metadata data
    :return: Dictionary of metadata (values as bytes)
    :raises SerializationError: If deserialization fails
    """
    try:
        pb_metadata = PeerMetadata()
        pb_metadata.ParseFromString(data)
        return dict(pb_metadata.metadata)
    except Exception as e:
        raise SerializationError(f"Failed to deserialize metadata: {e}") from e


def serialize_keypair(keypair: KeyPair) -> bytes:
    """
    Serialize a keypair to bytes.

    :param keypair: KeyPair to serialize
    :return: Serialized bytes
    :raises SerializationError: If serialization fails
    """
    try:
        pb_keys = PeerKeys()

        # Serialize public key
        if keypair.public_key:
            pb_public_key = PBPublicKey()

            key_type = keypair.public_key.get_type()
            # Map from libp2p KeyType to protobuf KeyType
            if key_type.value == 0:  # RSA
                pb_public_key.Type = PBKeyType.RSA
            elif key_type.value == 1:  # Ed25519
                pb_public_key.Type = PBKeyType.Ed25519
            elif key_type.value == 2:  # Secp256k1
                pb_public_key.Type = PBKeyType.Secp256k1
            elif key_type.value == 3:  # ECDSA
                pb_public_key.Type = PBKeyType.ECDSA
            else:
                raise SerializationError(f"Unsupported key type: {key_type}")
            pb_public_key.Data = keypair.public_key.serialize()
            pb_keys.public_key.CopyFrom(pb_public_key)

        # Serialize private key
        if keypair.private_key:
            pb_private_key = PBPrivateKey()

            key_type = keypair.private_key.get_type()
            # Map from libp2p KeyType to protobuf KeyType
            if key_type.value == 0:  # RSA
                pb_private_key.Type = PBKeyType.RSA
            elif key_type.value == 1:  # Ed25519
                pb_private_key.Type = PBKeyType.Ed25519
            elif key_type.value == 2:  # Secp256k1
                pb_private_key.Type = PBKeyType.Secp256k1
            elif key_type.value == 3:  # ECDSA
                pb_private_key.Type = PBKeyType.ECDSA
            else:
                raise SerializationError(f"Unsupported key type: {key_type}")
            pb_private_key.Data = keypair.private_key.serialize()
            pb_keys.private_key.CopyFrom(pb_private_key)

        return pb_keys.SerializeToString()
    except Exception as e:
        raise SerializationError(f"Failed to serialize keypair: {e}") from e


def deserialize_keypair(data: bytes) -> KeyPair:
    """
    Deserialize a keypair from bytes.

    :param data: Serialized keypair data
    :return: KeyPair object
    :raises SerializationError: If deserialization fails
    """
    try:
        pb_keys = PeerKeys()
        pb_keys.ParseFromString(data)

        from nanolib.crypto.serialization import (
            deserialize_private_key,
            deserialize_public_key,
        )

        public_key = None
        private_key = None

        if pb_keys.HasField("public_key"):
            public_key = deserialize_public_key(pb_keys.public_key.Data)

        if pb_keys.HasField("private_key"):
            private_key = deserialize_private_key(pb_keys.private_key.Data)

        # KeyPair requires both keys to be non-None
        if private_key is not None and public_key is not None:
            return KeyPair(private_key, public_key)
        elif private_key is not None:
            # If we only have private key, derive public key
            return KeyPair(private_key, private_key.get_public_key())
        else:
            # We can't create a KeyPair with only public key
            raise SerializationError("Cannot create KeyPair with only public key")
    except Exception as e:
        raise SerializationError(f"Failed to deserialize keypair: {e}") from e


def serialize_latency(latency_ns: int) -> bytes:
    """
    Serialize latency to bytes.

    :param latency_ns: Latency in nanoseconds
    :return: Serialized bytes
    :raises SerializationError: If serialization fails
    """
    try:
        pb_latency = PeerLatency()
        pb_latency.latency_ns = latency_ns
        return pb_latency.SerializeToString()
    except Exception as e:
        raise SerializationError(f"Failed to serialize latency: {e}") from e


def deserialize_latency(data: bytes) -> int:
    """
    Deserialize latency from bytes.

    :param data: Serialized latency data
    :return: Latency in nanoseconds
    :raises SerializationError: If deserialization fails
    """
    try:
        pb_latency = PeerLatency()
        pb_latency.ParseFromString(data)
        return pb_latency.latency_ns
    except Exception as e:
        raise SerializationError(f"Failed to deserialize latency: {e}") from e


def serialize_envelope(envelope: Envelope) -> bytes:
    """
    Serialize an envelope to bytes.

    :param envelope: Envelope to serialize
    :return: Serialized bytes
    :raises SerializationError: If serialization fails
    """
    try:
        pb_envelope_wrapper = PeerEnvelope()

        # Use the existing envelope's marshal_envelope method if available
        if hasattr(envelope, "marshal_envelope"):
            envelope_bytes = envelope.marshal_envelope()
            from nanolib.peer.pb.envelope_pb2 import Envelope as PBEnvelope

            pb_envelope = PBEnvelope()
            pb_envelope.ParseFromString(envelope_bytes)
            pb_envelope_wrapper.envelope.CopyFrom(pb_envelope)
        else:
            # Fallback: construct envelope manually
            from nanolib.peer.pb.envelope_pb2 import Envelope as PBEnvelope

            pb_envelope = PBEnvelope()
            pb_envelope.payload_type = envelope.payload_type
            pb_envelope.payload = envelope.raw_payload
            pb_envelope.signature = envelope.signature
            if envelope.public_key:
                # Convert the public key to protobuf format
                from nanolib.peer.envelope import pub_key_to_protobuf

                pb_envelope.public_key.CopyFrom(
                    pub_key_to_protobuf(envelope.public_key)
                )
            pb_envelope_wrapper.envelope.CopyFrom(pb_envelope)

        return pb_envelope_wrapper.SerializeToString()
    except Exception as e:
        raise SerializationError(f"Failed to serialize envelope: {e}") from e


def deserialize_envelope(data: bytes) -> Envelope:
    """
    Deserialize an envelope from bytes.

    :param data: Serialized envelope data
    :return: Envelope object
    :raises SerializationError: If deserialization fails
    """
    try:
        pb_envelope_wrapper = PeerEnvelope()
        pb_envelope_wrapper.ParseFromString(data)

        # Construct envelope manually from protobuf data
        pb_envelope = pb_envelope_wrapper.envelope

        # Convert protobuf public key back to crypto public key
        from nanolib.crypto.serialization import deserialize_public_key

        public_key_data = pb_envelope.public_key.SerializeToString()
        public_key = deserialize_public_key(public_key_data)

        return Envelope(
            public_key=public_key,
            payload_type=pb_envelope.payload_type,
            raw_payload=pb_envelope.payload,
            signature=pb_envelope.signature,
        )
    except Exception as e:
        raise SerializationError(f"Failed to deserialize envelope: {e}") from e


def serialize_record_state(state: "PeerRecordState") -> bytes:
    """
    Serialize a peer record state to bytes.

    :param state: PeerRecordState to serialize
    :return: Serialized bytes
    :raises SerializationError: If serialization fails
    """
    try:
        pb_state = PBPeerRecordState()

        # PeerRecordState is a simple class with envelope and seq
        # For now, we'll just mark it as VALID since we don't have state info
        # In the future, this could be extended to track actual state
        pb_state.state = PBPeerRecordState.State.VALID
        return pb_state.SerializeToString()
    except Exception as e:
        raise SerializationError(f"Failed to serialize record state: {e}") from e


def deserialize_record_state(data: bytes) -> "PeerRecordState":
    """
    Deserialize a peer record state from bytes.

    :param data: Serialized record state data
    :return: PeerRecordState
    :raises SerializationError: If deserialization fails
    """
    try:
        pb_state = PBPeerRecordState()
        pb_state.ParseFromString(data)

        # Since we can't reconstruct the full PeerRecordState without the envelope
        # and sequence number (seq), we'll need to return a placeholder. This is a
        # limitation of the current design. In practice, the record state should
        # be stored together with its envelope and seq.
        from nanolib.crypto.ed25519 import Ed25519PublicKey
        from nanolib.peer.envelope import Envelope
        from nanolib.peer.peerstore import PeerRecordState

        # Create a dummy envelope for now
        dummy_key = Ed25519PublicKey.from_bytes(b"\x00" * 32)
        dummy_envelope = Envelope(
            public_key=dummy_key, payload_type=b"", raw_payload=b"", signature=b""
        )

        return PeerRecordState(dummy_envelope, 0)
    except Exception as e:
        raise SerializationError(f"Failed to deserialize record state: {e}") from e



from nanolib.exceptions import (
    BaseLibp2pError,
)


class HostException(BaseLibp2pError):
    """A generic exception  in `IHost`."""


class ConnectionFailure(HostException):
    pass


class StreamFailure(HostException):
    pass



from collections.abc import Sequence
import threading
import time
from typing import Any

from multiaddr import Multiaddr

from nanolib.abc import IPeerRecord
from nanolib.peer.id import ID
import nanolib.peer.pb.peer_record_pb2 as pb
from nanolib.peer.peerinfo import PeerInfo

PEER_RECORD_ENVELOPE_DOMAIN = "libp2p-peer-record"
PEER_RECORD_ENVELOPE_PAYLOAD_TYPE = b"\x03\x01"

_last_timestamp_lock = threading.Lock()
_last_timestamp: int = 0


class PeerRecord(IPeerRecord):
    """
    A record that contains metatdata about a peer in the libp2p network.

    This includes:
    - `peer_id`: The peer's globally unique indentifier.
    - `addrs`: A list of the peer's publicly reachable multiaddrs.
    - `seq`: A strictly monotonically increasing timestamp used
    to order records over time.

    PeerRecords are designed to be signed and transmitted in libp2p routing Envelopes.
    """

    peer_id: ID
    addrs: list[Multiaddr]
    seq: int

    def __init__(
        self,
        peer_id: ID | None = None,
        addrs: list[Multiaddr] | None = None,
        seq: int | None = None,
    ) -> None:
        """
        Initialize a new PeerRecord.
        If `seq` is not provided, a timestamp-based strictly increasing sequence
        number will be generated.

        :param peer_id: ID of the peer this record refers to.
        :param addrs: Public multiaddrs of the peer.
        :param seq: Monotonic sequence number.

        """
        if peer_id is not None:
            self.peer_id = peer_id
        self.addrs = addrs or []
        if seq is not None:
            self.seq = seq
        else:
            self.seq = timestamp_seq()

    def __repr__(self) -> str:
        return (
            f"PeerRecord(\n"
            f"  peer_id={self.peer_id},\n"
            f"  multiaddrs={[str(m) for m in self.addrs]},\n"
            f"  seq={self.seq}\n"
            f")"
        )

    def domain(self) -> str:
        """
        Return the domain string associated with this PeerRecord.

        Used during record signing and envelope validation to identify the record type.
        """
        return PEER_RECORD_ENVELOPE_DOMAIN

    def codec(self) -> bytes:
        """
        Return the codec identifier for PeerRecords.

        This binary perfix helps distinguish PeerRecords in serialized envelopes.
        """
        return PEER_RECORD_ENVELOPE_PAYLOAD_TYPE

    def to_protobuf(self) -> pb.PeerRecord:
        """
        Convert the current PeerRecord into a ProtoBuf PeerRecord message.

        :raises ValueError: if peer_id serialization fails.
        :return: A ProtoBuf-encoded PeerRecord message object.
        """
        try:
            id_bytes = self.peer_id.to_bytes()
        except Exception as e:
            raise ValueError(f"failed to marshal peer_id: {e}")

        msg = pb.PeerRecord()
        msg.peer_id = id_bytes
        msg.seq = self.seq
        msg.addresses.extend(addrs_to_protobuf(self.addrs))
        return msg

    def marshal_record(self) -> bytes:
        """
        Serialize a PeerRecord into raw bytes suitable for embedding in an Envelope.

        This is typically called during the process of signing or sealing the record.
        :raises ValueError: if serialization to protobuf fails.
        :return: Serialized PeerRecord bytes.
        """
        try:
            msg = self.to_protobuf()
            return msg.SerializeToString()
        except Exception as e:
            raise ValueError(f"failed to marshal PeerRecord: {e}")

    def equal(self, other: Any) -> bool:
        """
        Check if this PeerRecord is identical to another.

        Two PeerRecords are considered equal if:
        - Their peer IDs match.
        - Their sequence numbers are identical.
        - Their address lists are identical and in the same order.

        :param other: Another PeerRecord instance.
        :return: True if all fields mathch, False otherwise.
        """
        if isinstance(other, PeerRecord):
            if self.peer_id == other.peer_id:
                if self.seq == other.seq:
                    if len(self.addrs) == len(other.addrs):
                        for a1, a2 in zip(self.addrs, other.addrs):
                            if a1 == a2:
                                continue
                            else:
                                return False
                        return True
        return False


def unmarshal_record(data: bytes) -> PeerRecord:
    """
    Deserialize a PeerRecord from its serialized byte representation.

    Typically used when receiveing a PeerRecord inside a signed routing Envelope.

    :param data: Serialized protobuf-encoded bytes.
    :raises ValueError: if parsing or conversion fails.
    :reurn: A valid PeerRecord instance.
    """
    if data is None:
        raise ValueError("cannot unmarshal PeerRecord from None")

    msg = pb.PeerRecord()
    try:
        msg.ParseFromString(data)
    except Exception as e:
        raise ValueError(f"Failed to parse PeerRecord protobuf: {e}")

    try:
        record = peer_record_from_protobuf(msg)
    except Exception as e:
        raise ValueError(f"Failed to convert protobuf to PeerRecord: {e}")

    return record


def timestamp_seq() -> int:
    """
    Generate a strictly increasing timestamp-based sequence number.

    Ensures that even if multiple PeerRecords are generated in the same nanosecond,
    their `seq` values will still be strictly increasing by using a lock to track the
    last value.

    :return: A strictly increasing integer timestamp.
    """
    global _last_timestamp
    now = int(time.time_ns())
    with _last_timestamp_lock:
        if now <= _last_timestamp:
            now = _last_timestamp + 1
        _last_timestamp = now
    return now


def peer_record_from_peer_info(info: PeerInfo) -> PeerRecord:
    """
    Create a PeerRecord from a PeerInfo object.

    This automatically assigns a timestamp-based sequence number to the record.
    :param info: A PeerInfo instance (contains peer_id and addrs).
    :return: A PeerRecord instance.
    """
    record = PeerRecord()
    record.peer_id = info.peer_id
    record.addrs = info.addrs
    return record


def peer_record_from_protobuf(msg: pb.PeerRecord) -> PeerRecord:
    """
    Convert a protobuf PeerRecord message into a PeerRecord object.

    :param msg: Protobuf PeerRecord message.
    :raises ValueError: if the peer_id cannot be parsed.
    :return: A deserialized PeerRecord instance.
    """
    try:
        peer_id = ID(msg.peer_id)
    except Exception as e:
        raise ValueError(f"Failed to unmarshal peer_id: {e}")

    addrs = addrs_from_protobuf(msg.addresses)
    seq = msg.seq

    return PeerRecord(peer_id, addrs, seq)


def addrs_from_protobuf(addrs: Sequence[pb.PeerRecord.AddressInfo]) -> list[Multiaddr]:
    """
    Convert a list of protobuf address records to Multiaddr objects.

    :param addrs: A list of protobuf PeerRecord.AddressInfo messages.
    :return: A list of decoded Multiaddr instances (invalid ones are skipped).
    """
    out = []
    for addr_info in addrs:
        try:
            addr = Multiaddr(addr_info.multiaddr)
            out.append(addr)
        except Exception:
            continue
    return out


def addrs_to_protobuf(addrs: list[Multiaddr]) -> list[pb.PeerRecord.AddressInfo]:
    """
    Convert a list of Multiaddr objects into their protobuf representation.

    :param addrs: A list of Multiaddr instances.
    :return: A list of PeerRecord.AddressInfo protobuf messages.
    """
    out = []
    for addr in addrs:
        addr_info = pb.PeerRecord.AddressInfo()
        addr_info.multiaddr = addr.to_bytes()
        out.append(addr_info)
    return out



import functools
import hashlib

import base58
import multihash

from nanolib.crypto.keys import (
    PublicKey,
)

# NOTE: On inlining...
# See: https://github.com/libp2p/specs/issues/138
# NOTE: enabling to be interoperable w/ the Go implementation
ENABLE_INLINING = True
MAX_INLINE_KEY_LENGTH = 42

IDENTITY_MULTIHASH_CODE = 0x00

if ENABLE_INLINING:

    class IdentityHash:
        _digest: bytes

        def __init__(self) -> None:
            self._digest = b""

        def update(self, input: bytes) -> None:
            self._digest += input

        def digest(self) -> bytes:
            return self._digest

    # Register identity hash function if FuncReg is available
    if hasattr(multihash, "FuncReg"):
        multihash.FuncReg.register(
            IDENTITY_MULTIHASH_CODE, "identity", hash_new=lambda: IdentityHash()
        )


class ID:
    _bytes: bytes

    def __init__(self, peer_id_bytes: bytes) -> None:
        self._bytes = peer_id_bytes

    @functools.cached_property
    def xor_id(self) -> int:
        return int(sha256_digest(self._bytes).hex(), 16)

    @functools.cached_property
    def base58(self) -> str:
        return base58.b58encode(self._bytes).decode()

    def to_bytes(self) -> bytes:
        return self._bytes

    def to_base58(self) -> str:
        return self.base58

    def __repr__(self) -> str:
        return f"<libp2p.peer.id.ID ({self!s})>"

    __str__ = pretty = to_string = to_base58

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.to_base58() == other
        elif isinstance(other, bytes):
            return self._bytes == other
        elif isinstance(other, ID):
            return self._bytes == other._bytes
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(self._bytes)

    @classmethod
    def from_base58(cls, b58_encoded_peer_id_str: str) -> "ID":
        peer_id_bytes = base58.b58decode(b58_encoded_peer_id_str)
        pid = ID(peer_id_bytes)
        return pid

    @classmethod
    def from_pubkey(cls, key: PublicKey) -> "ID":
        serialized_key = key.serialize()
        algo = multihash.Func.sha2_256
        if ENABLE_INLINING and len(serialized_key) <= MAX_INLINE_KEY_LENGTH:
            algo = IDENTITY_MULTIHASH_CODE
        mh_digest = multihash.digest(serialized_key, algo)
        return cls(mh_digest.encode())


def sha256_digest(data: str | bytes) -> bytes:
    if isinstance(data, str):
        data = data.encode("utf8")
    return hashlib.sha256(data).digest()



from collections import (
    OrderedDict,
)

import trio

from nanolib.abc import (
    IMuxedConn,
    IRawConnection,
    ISecureConn,
)
from nanolib.custom_types import (
    TMuxerClass,
    TMuxerOptions,
    TProtocol,
)
from nanolib.peer.id import (
    ID,
)
from nanolib.protocol_muxer.exceptions import (
    MultiselectError,
)
from nanolib.protocol_muxer.multiselect import (
    DEFAULT_NEGOTIATE_TIMEOUT,
    Multiselect,
)
from nanolib.protocol_muxer.multiselect_client import (
    MultiselectClient,
)
from nanolib.protocol_muxer.multiselect_communicator import (
    MultiselectCommunicator,
)
from nanolib.stream_muxer.yamux.yamux import (
    PROTOCOL_ID,
    Yamux,
)


class MuxerMultistream:
    """
    MuxerMultistream is a multistream stream muxed transport multiplexer.

    go implementation: github.com/libp2p/go-stream-muxer-multistream/multistream.go
    """

    # NOTE: Can be changed to `typing.OrderedDict` since Python 3.7.2.
    transports: "OrderedDict[TProtocol, TMuxerClass]"
    multiselect: Multiselect
    multiselect_client: MultiselectClient
    negotiate_timeout: int

    def __init__(
        self,
        muxer_transports_by_protocol: TMuxerOptions,
        negotiate_timeout: int = DEFAULT_NEGOTIATE_TIMEOUT,
    ) -> None:
        self.transports = OrderedDict()
        self.multiselect = Multiselect()
        self.multistream_client = MultiselectClient()
        self.negotiate_timeout = negotiate_timeout
        for protocol, transport in muxer_transports_by_protocol.items():
            self.add_transport(protocol, transport)

    def add_transport(self, protocol: TProtocol, transport: TMuxerClass) -> None:
        """
        Add a protocol and its corresponding transport to multistream-
        select(multiselect). The order that a protocol is added is exactly the
        precedence it is negotiated in multiselect.

        :param protocol: the protocol name, which is negotiated in multiselect.
        :param transport: the corresponding transportation to the ``protocol``.
        """
        # If protocol is already added before, remove it and add it again.
        self.transports.pop(protocol, None)
        self.transports[protocol] = transport
        self.multiselect.add_handler(protocol, None)

    async def select_transport(self, conn: IRawConnection) -> TMuxerClass:
        """
        Select a transport that both us and the node on the other end of conn
        support and agree on.

        :param conn: conn to choose a transport over
        :return: selected muxer transport
        """
        protocol: TProtocol | None
        communicator = MultiselectCommunicator(conn)
        if conn.is_initiator:
            protocol = await self.multiselect_client.select_one_of(
                tuple(self.transports.keys()), communicator, self.negotiate_timeout
            )
        else:
            protocol, _ = await self.multiselect.negotiate(
                communicator, self.negotiate_timeout
            )
        if protocol is None:
            raise MultiselectError(
                "Fail to negotiate a stream muxer protocol: no protocol selected"
            )
        return self.transports[protocol]

    async def new_conn(self, conn: ISecureConn, peer_id: ID) -> IMuxedConn:
        communicator = MultiselectCommunicator(conn)
        protocol = await self.multistream_client.select_one_of(
            tuple(self.transports.keys()), communicator, self.negotiate_timeout
        )
        transport_class = self.transports[protocol]
        if protocol == PROTOCOL_ID:
            async with trio.open_nursery():

                async def on_close() -> None:
                    pass

                return Yamux(
                    conn, peer_id, is_initiator=conn.is_initiator, on_close=on_close
                )
        return transport_class(conn, peer_id)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
from typing import Any, Optional

from . import autonat_pb2 as autonat__pb2

GRPC_GENERATED_VERSION = "1.71.0"
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(
        GRPC_VERSION, GRPC_GENERATED_VERSION
    )
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + f" but the generated code in autonat_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
    )


class AutoNATStub:
    """AutoNAT service definition"""

    def __init__(self, channel: grpc.Channel) -> None:
        """
        Initialize the AutoNAT stub.

        Args:
        ----
        channel (grpc.Channel): The gRPC channel instance that facilitates
            communication for the AutoNAT service, providing the underlying
            transport mechanism for RPC calls.

        """
        self.Dial = channel.unary_unary(
            "/autonat.pb.AutoNAT/Dial",
            request_serializer=autonat__pb2.Message.SerializeToString,
            response_deserializer=autonat__pb2.Message.FromString,
            _registered_method=True,
        )


class AutoNATServicer:
    """AutoNAT service definition"""

    def Dial(self, request: Any, context: Any) -> Any:
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_AutoNATServicer_to_server(servicer: AutoNATServicer, server: Any) -> None:
    rpc_method_handlers = {
        "Dial": grpc.unary_unary_rpc_method_handler(
            servicer.Dial,
            request_deserializer=autonat__pb2.Message.FromString,
            response_serializer=autonat__pb2.Message.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "autonat.pb.AutoNAT", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers("autonat.pb.AutoNAT", rpc_method_handlers)


# This class is part of an EXPERIMENTAL API.
class AutoNAT:
    """AutoNAT service definition"""

    @staticmethod
    def Dial(
        request: Any,
        target: str,
        options: tuple[Any, ...] = (),
        channel_credentials: Any | None = None,
        call_credentials: Any | None = None,
        insecure: bool = False,
        compression: Any | None = None,
        wait_for_ready: bool | None = None,
        timeout: float | None = None,
        metadata: list[tuple[str, str]] | None = None,
    ) -> Any:
        channel = grpc.secure_channel(target, channel_credentials) if channel_credentials else grpc.insecure_channel(target)
        return channel.unary_unary(
            "/autonat.pb.AutoNAT/Dial",
            request_serializer=autonat__pb2.Message.SerializeToString,
            response_deserializer=autonat__pb2.Message.FromString,
            _registered_method=True,
        )(
            request,
            timeout=timeout,
            metadata=metadata,
            wait_for_ready=wait_for_ready,
        )



"""
In-memory datastore implementation.

This provides a simple in-memory datastore for testing and development.
"""

from collections.abc import Iterator

import trio

from .base import IBatch, IBatchingDatastore


class MemoryBatch(IBatch):
    """In-memory batch implementation."""

    def __init__(self, datastore: "MemoryDatastore"):
        self.datastore = datastore
        self.operations: list[tuple[str, bytes, bytes | None]] = []

    async def put(self, key: bytes, value: bytes) -> None:
        """Add a put operation to the batch."""
        self.operations.append(("put", key, value))

    async def delete(self, key: bytes) -> None:
        """Add a delete operation to the batch."""
        self.operations.append(("delete", key, None))

    async def commit(self) -> None:
        """Commit all operations in the batch."""
        for op_type, key, value in self.operations:
            if op_type == "put":
                if value is None:
                    raise ValueError("Cannot put None value in memory datastore")
                self.datastore._data[key] = value
            elif op_type == "delete":
                self.datastore._data.pop(key, None)
        self.operations.clear()


class MemoryDatastore(IBatchingDatastore):
    """
    In-memory datastore implementation.

    This is useful for testing and development scenarios where persistence
    is not required.
    """

    def __init__(self) -> None:
        self._data: dict[bytes, bytes] = {}
        self._lock = trio.Lock()

    async def get(self, key: bytes) -> bytes | None:
        """Retrieve a value by key."""
        async with self._lock:
            return self._data.get(key)

    async def put(self, key: bytes, value: bytes) -> None:
        """Store a key-value pair."""
        async with self._lock:
            self._data[key] = value

    async def delete(self, key: bytes) -> None:
        """Delete a key-value pair."""
        async with self._lock:
            self._data.pop(key, None)

    async def has(self, key: bytes) -> bool:
        """Check if a key exists."""
        async with self._lock:
            return key in self._data

    def query(self, prefix: bytes) -> Iterator[tuple[bytes, bytes]]:
        """Query for keys with a given prefix."""
        # Note: This is not thread-safe, but for in-memory datastore it's acceptable
        for key, value in self._data.items():
            if key.startswith(prefix):
                yield (key, value)

    async def close(self) -> None:
        """Close the datastore and clean up resources."""
        async with self._lock:
            self._data.clear()

    async def sync(self, prefix: bytes) -> None:
        """Sync any pending writes to disk (no-op for in-memory)."""
        pass

    async def batch(self) -> MemoryBatch:
        """Create a new batch for batched operations."""
        return MemoryBatch(self)



"""
Security modules for libp2p.

This package provides various security implementations including:
- TLS transport
- Noise protocol
- SECIO protocol
- Insecure transport (for testing)
"""

# Import TLS module
from . import tls

# Import other security modules
from . import insecure
from . import noise
from . import secio

__all__ = [
    "tls",
    "insecure",
    "noise",
    "secio",
]



"""
NanoPy - Ethereum-compatible blockchain fork
"""

__version__ = "1.2.78"
__chain_id__ = 1337

from nanopy.core import Block, Transaction, Account, WorldState
from nanopy.crypto import generate_keypair, Wallet

__all__ = [
    "Block",
    "Transaction",
    "Account",
    "WorldState",
    "generate_keypair",
    "Wallet",
    "__version__",
    "__chain_id__",
]



from .abc import (
    ServiceAPI,
)
from .base import (
    Service,
    as_service,
)
from .exceptions import (
    DaemonTaskExit,
    LifecycleError,
)
from .trio_service import (
    TrioManager,
    background_trio_service,
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: libp2p/discovery/rendezvous/pb/rendezvous.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/libp2p/discovery/rendezvous/pb/rendezvous.proto\x12\rrendezvous.pb\"\xa8\x0c\n\x07Message\x12\x30\n\x04type\x18\x01 \x01(\x0e\x32\".rendezvous.pb.Message.MessageType\x12\x31\n\x08register\x18\x02 \x01(\x0b\x32\x1f.rendezvous.pb.Message.Register\x12\x41\n\x10registerResponse\x18\x03 \x01(\x0b\x32\'.rendezvous.pb.Message.RegisterResponse\x12\x35\n\nunregister\x18\x04 \x01(\x0b\x32!.rendezvous.pb.Message.Unregister\x12\x31\n\x08\x64iscover\x18\x05 \x01(\x0b\x32\x1f.rendezvous.pb.Message.Discover\x12\x41\n\x10\x64iscoverResponse\x18\x06 \x01(\x0b\x32\'.rendezvous.pb.Message.DiscoverResponse\x12\x43\n\x11\x64iscoverSubscribe\x18\x64 \x01(\x0b\x32(.rendezvous.pb.Message.DiscoverSubscribe\x12S\n\x19\x64iscoverSubscribeResponse\x18\x65 \x01(\x0b\x32\x30.rendezvous.pb.Message.DiscoverSubscribeResponse\x1a%\n\x08PeerInfo\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\r\n\x05\x61\x64\x64rs\x18\x02 \x03(\x0c\x1aR\n\x08Register\x12\n\n\x02ns\x18\x01 \x01(\t\x12-\n\x04peer\x18\x02 \x01(\x0b\x32\x1f.rendezvous.pb.Message.PeerInfo\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\x1aj\n\x10RegisterResponse\x12\x35\n\x06status\x18\x01 \x01(\x0e\x32%.rendezvous.pb.Message.ResponseStatus\x12\x12\n\nstatusText\x18\x02 \x01(\t\x12\x0b\n\x03ttl\x18\x03 \x01(\x03\x1a$\n\nUnregister\x12\n\n\x02ns\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\x0c\x1a\x35\n\x08\x44iscover\x12\n\n\x02ns\x18\x01 \x01(\t\x12\r\n\x05limit\x18\x02 \x01(\x03\x12\x0e\n\x06\x63ookie\x18\x03 \x01(\x0c\x1a\xa5\x01\n\x10\x44iscoverResponse\x12\x36\n\rregistrations\x18\x01 \x03(\x0b\x32\x1f.rendezvous.pb.Message.Register\x12\x0e\n\x06\x63ookie\x18\x02 \x01(\x0c\x12\x35\n\x06status\x18\x03 \x01(\x0e\x32%.rendezvous.pb.Message.ResponseStatus\x12\x12\n\nstatusText\x18\x04 \x01(\t\x1a\x45\n\x11\x44iscoverSubscribe\x12$\n\x1csupported_subscription_types\x18\x01 \x03(\t\x12\n\n\x02ns\x18\x02 \x01(\t\x1a\x9f\x01\n\x19\x44iscoverSubscribeResponse\x12\x19\n\x11subscription_type\x18\x01 \x01(\t\x12\x1c\n\x14subscription_details\x18\x02 \x01(\t\x12\x35\n\x06status\x18\x03 \x01(\x0e\x32%.rendezvous.pb.Message.ResponseStatus\x12\x12\n\nstatusText\x18\x04 \x01(\t\"\xa0\x01\n\x0bMessageType\x12\x0c\n\x08REGISTER\x10\x00\x12\x15\n\x11REGISTER_RESPONSE\x10\x01\x12\x0e\n\nUNREGISTER\x10\x02\x12\x0c\n\x08\x44ISCOVER\x10\x03\x12\x15\n\x11\x44ISCOVER_RESPONSE\x10\x04\x12\x16\n\x12\x44ISCOVER_SUBSCRIBE\x10\x64\x12\x1f\n\x1b\x44ISCOVER_SUBSCRIBE_RESPONSE\x10\x65\"\xb5\x01\n\x0eResponseStatus\x12\x06\n\x02OK\x10\x00\x12\x17\n\x13\x45_INVALID_NAMESPACE\x10\x64\x12\x17\n\x13\x45_INVALID_PEER_INFO\x10\x65\x12\x11\n\rE_INVALID_TTL\x10\x66\x12\x14\n\x10\x45_INVALID_COOKIE\x10g\x12\x15\n\x10\x45_NOT_AUTHORIZED\x10\xc8\x01\x12\x15\n\x10\x45_INTERNAL_ERROR\x10\xac\x02\x12\x12\n\rE_UNAVAILABLE\x10\x90\x03\"H\n\x12RegistrationRecord\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05\x61\x64\x64rs\x18\x02 \x03(\x0c\x12\n\n\x02ns\x18\x03 \x01(\t\x12\x0b\n\x03ttl\x18\x04 \x01(\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nanolib.discovery.rendezvous.pb.rendezvous_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_MESSAGE']._serialized_start=67
  _globals['_MESSAGE']._serialized_end=1643
  _globals['_MESSAGE_PEERINFO']._serialized_start=573
  _globals['_MESSAGE_PEERINFO']._serialized_end=610
  _globals['_MESSAGE_REGISTER']._serialized_start=612
  _globals['_MESSAGE_REGISTER']._serialized_end=694
  _globals['_MESSAGE_REGISTERRESPONSE']._serialized_start=696
  _globals['_MESSAGE_REGISTERRESPONSE']._serialized_end=802
  _globals['_MESSAGE_UNREGISTER']._serialized_start=804
  _globals['_MESSAGE_UNREGISTER']._serialized_end=840
  _globals['_MESSAGE_DISCOVER']._serialized_start=842
  _globals['_MESSAGE_DISCOVER']._serialized_end=895
  _globals['_MESSAGE_DISCOVERRESPONSE']._serialized_start=898
  _globals['_MESSAGE_DISCOVERRESPONSE']._serialized_end=1063
  _globals['_MESSAGE_DISCOVERSUBSCRIBE']._serialized_start=1065
  _globals['_MESSAGE_DISCOVERSUBSCRIBE']._serialized_end=1134
  _globals['_MESSAGE_DISCOVERSUBSCRIBERESPONSE']._serialized_start=1137
  _globals['_MESSAGE_DISCOVERSUBSCRIBERESPONSE']._serialized_end=1296
  _globals['_MESSAGE_MESSAGETYPE']._serialized_start=1299
  _globals['_MESSAGE_MESSAGETYPE']._serialized_end=1459
  _globals['_MESSAGE_RESPONSESTATUS']._serialized_start=1462
  _globals['_MESSAGE_RESPONSESTATUS']._serialized_end=1643
  _globals['_REGISTRATIONRECORD']._serialized_start=1645
  _globals['_REGISTRATIONRECORD']._serialized_end=1717
# @@protoc_insertion_point(module_scope)



from dataclasses import (
    dataclass,
    field,
)
import logging

from nanolib.crypto.keys import (
    PrivateKey,
    PublicKey,
)
from nanolib.crypto.serialization import (
    deserialize_public_key,
)

from .pb import noise_pb2 as noise_pb

logger = logging.getLogger(__name__)

SIGNED_DATA_PREFIX = "noise-libp2p-static-key:"


@dataclass
class NoiseExtensions:
    """
    Noise protocol extensions for advanced features like WebTransport and early data.

    This class provides support for:
    - WebTransport certificate hashes for WebTransport support
    - Stream multiplexers supported by this peer (spec compliant)
    - Early data payload for 0-RTT support (Python extension)
    """

    webtransport_certhashes: list[bytes] = field(default_factory=list)
    stream_muxers: list[str] = field(default_factory=list)
    early_data: bytes | None = None

    def to_protobuf(self) -> noise_pb.NoiseExtensions:
        """
        Convert to protobuf message.

        Returns:
            noise_pb.NoiseExtensions: The protobuf message representation

        """
        ext = noise_pb.NoiseExtensions()
        ext.webtransport_certhashes.extend(self.webtransport_certhashes)
        ext.stream_muxers.extend(self.stream_muxers)  # type: ignore[attr-defined]
        if self.early_data is not None:
            ext.early_data = self.early_data
        return ext

    @classmethod
    def from_protobuf(cls, pb_ext: noise_pb.NoiseExtensions) -> "NoiseExtensions":
        """
        Create from protobuf message.

        Args:
            pb_ext: The protobuf message to convert

        Returns:
            NoiseExtensions: The Python dataclass representation

        """
        early_data = None
        if pb_ext.HasField("early_data"):
            early_data = pb_ext.early_data
        return cls(
            webtransport_certhashes=list(pb_ext.webtransport_certhashes),
            stream_muxers=list(pb_ext.stream_muxers),  # type: ignore[attr-defined]
            early_data=early_data,
        )

    def is_empty(self) -> bool:
        """
        Check if extensions are empty (no data).

        Returns:
            bool: True if no extensions data is present

        """
        return (
            not self.webtransport_certhashes
            and not self.stream_muxers
            and self.early_data is None
        )

    def has_webtransport_certhashes(self) -> bool:
        """
        Check if WebTransport certificate hashes are present.

        Returns:
            bool: True if WebTransport certificate hashes are present

        """
        return bool(self.webtransport_certhashes)

    def has_stream_muxers(self) -> bool:
        """
        Check if stream multiplexers are present.

        Returns:
            bool: True if stream multiplexers are present

        """
        return bool(self.stream_muxers)

    def has_early_data(self) -> bool:
        """
        Check if early data is present.

        Returns:
            bool: True if early data is present

        """
        return self.early_data is not None


@dataclass
class NoiseHandshakePayload:
    """
    Noise handshake payload containing peer identity and optional extensions.

    This class represents the payload sent during Noise handshake and provides:
    - Peer identity verification through public key and signature
    - Optional extensions for advanced features like WebTransport and stream muxers
    """

    id_pubkey: PublicKey
    id_sig: bytes
    extensions: NoiseExtensions | None = None

    def serialize(self) -> bytes:
        """
        Serialize the handshake payload to protobuf bytes.

        Returns:
            bytes: The serialized protobuf message

        Raises:
            ValueError: If the payload is invalid

        """
        if not self.id_pubkey or not self.id_sig:
            raise ValueError("Invalid handshake payload: missing required fields")

        msg = noise_pb.NoiseHandshakePayload(
            identity_key=self.id_pubkey.serialize(), identity_sig=self.id_sig
        )

        # Include extensions if present
        if self.extensions is not None:
            msg.extensions.CopyFrom(self.extensions.to_protobuf())

        return msg.SerializeToString()

    @classmethod
    def deserialize(cls, protobuf_bytes: bytes) -> "NoiseHandshakePayload":
        """
        Deserialize protobuf bytes to handshake payload.

        Args:
            protobuf_bytes: The serialized protobuf message

        Returns:
            NoiseHandshakePayload: The deserialized handshake payload

        Raises:
            ValueError: If the protobuf data is invalid

        """
        if not protobuf_bytes:
            raise ValueError("Empty protobuf data")

        try:
            msg = noise_pb.NoiseHandshakePayload.FromString(protobuf_bytes)
        except Exception as e:
            raise ValueError(f"Failed to deserialize protobuf: {e}")

        if not msg.identity_key or not msg.identity_sig:
            raise ValueError("Invalid handshake payload: missing required fields")

        extensions = None
        if msg.HasField("extensions"):
            extensions = NoiseExtensions.from_protobuf(msg.extensions)

        try:
            id_pubkey = deserialize_public_key(msg.identity_key)
        except Exception as e:
            raise ValueError(f"Failed to deserialize public key: {e}")

        return cls(
            id_pubkey=id_pubkey,
            id_sig=msg.identity_sig,
            extensions=extensions,
        )

    def has_extensions(self) -> bool:
        """
        Check if extensions are present.

        Returns:
            bool: True if extensions are present

        """
        return self.extensions is not None and not self.extensions.is_empty()

    def has_early_data(self) -> bool:
        """
        Check if early data is present in extensions.

        Returns:
            bool: True if early data is present

        """
        return self.extensions is not None and self.extensions.has_early_data()

    def get_early_data(self) -> bytes | None:
        """
        Get early data from extensions.

        Returns:
            bytes | None: The early data if present

        """
        if self.extensions is not None and self.extensions.has_early_data():
            return self.extensions.early_data
        return None


def make_data_to_be_signed(noise_static_pubkey: PublicKey) -> bytes:
    prefix_bytes = SIGNED_DATA_PREFIX.encode("utf-8")
    return prefix_bytes + noise_static_pubkey.to_bytes()


def make_handshake_payload_sig(
    id_privkey: PrivateKey, noise_static_pubkey: PublicKey
) -> bytes:
    data = make_data_to_be_signed(noise_static_pubkey)
    logger.debug(f"make_handshake_payload_sig: signing data length: {len(data)}")
    logger.debug(f"make_handshake_payload_sig: signing data hex: {data.hex()}")
    return id_privkey.sign(data)


def verify_handshake_payload_sig(
    payload: NoiseHandshakePayload, noise_static_pubkey: PublicKey
) -> bool:
    """
    Verify if the signature
        1. is composed of the data `SIGNED_DATA_PREFIX`++`noise_static_pubkey` and
        2. signed by the private key corresponding to `id_pubkey`
    """
    expected_data = make_data_to_be_signed(noise_static_pubkey)
    logger.debug(
        f"verify_handshake_payload_sig: payload.id_pubkey type: "
        f"{type(payload.id_pubkey)}"
    )
    logger.debug(
        f"verify_handshake_payload_sig: noise_static_pubkey type: "
        f"{type(noise_static_pubkey)}"
    )
    logger.debug(
        f"verify_handshake_payload_sig: expected_data length: {len(expected_data)}"
    )
    logger.debug(
        f"verify_handshake_payload_sig: expected_data hex: {expected_data.hex()}"
    )
    logger.debug(
        f"verify_handshake_payload_sig: payload.id_sig length: {len(payload.id_sig)}"
    )
    try:
        result = payload.id_pubkey.verify(expected_data, payload.id_sig)
        logger.debug(f"verify_handshake_payload_sig: verification result: {result}")
        return result
    except Exception as e:
        logger.error(f"verify_handshake_payload_sig: verification exception: {e}")
        return False



from nanolib.exceptions import (
    BaseLibp2pError,
)


class OpenConnectionError(BaseLibp2pError):
    pass


class UpgradeFailure(BaseLibp2pError):
    pass


class SecurityUpgradeFailure(UpgradeFailure):
    pass


class MuxerUpgradeFailure(UpgradeFailure):
    pass



from nanolib.abc import (
    IRawConnection,
)
from nanolib.io.abc import (
    ReadWriteCloser,
)
from nanolib.io.exceptions import (
    IOException,
)

from .exceptions import (
    RawConnError,
)


class RawConnection(IRawConnection):
    stream: ReadWriteCloser
    is_initiator: bool

    def __init__(self, stream: ReadWriteCloser, initiator: bool) -> None:
        self.stream = stream
        self.is_initiator = initiator

    async def write(self, data: bytes) -> None:
        """Raise `RawConnError` if the underlying connection breaks."""
        try:
            await self.stream.write(data)
        except IOException as error:
            raise RawConnError from error

    async def read(self, n: int | None = None) -> bytes:
        """
        Read up to ``n`` bytes from the underlying stream. This call is
        delegated directly to the underlying ``self.reader``.

        Raise `RawConnError` if the underlying connection breaks
        """
        try:
            return await self.stream.read(n)
        except IOException as error:
            raise RawConnError from error

    async def close(self) -> None:
        await self.stream.close()

    def get_remote_address(self) -> tuple[str, int] | None:
        """Delegate to the underlying stream's get_remote_address method."""
        return self.stream.get_remote_address()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Message(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _MessageType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _MessageTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Message._MessageType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        REGISTER: Message._MessageType.ValueType  # 0
        REGISTER_RESPONSE: Message._MessageType.ValueType  # 1
        UNREGISTER: Message._MessageType.ValueType  # 2
        DISCOVER: Message._MessageType.ValueType  # 3
        DISCOVER_RESPONSE: Message._MessageType.ValueType  # 4
        DISCOVER_SUBSCRIBE: Message._MessageType.ValueType  # 100
        DISCOVER_SUBSCRIBE_RESPONSE: Message._MessageType.ValueType  # 101

    class MessageType(_MessageType, metaclass=_MessageTypeEnumTypeWrapper): ...
    REGISTER: Message.MessageType.ValueType  # 0
    REGISTER_RESPONSE: Message.MessageType.ValueType  # 1
    UNREGISTER: Message.MessageType.ValueType  # 2
    DISCOVER: Message.MessageType.ValueType  # 3
    DISCOVER_RESPONSE: Message.MessageType.ValueType  # 4
    DISCOVER_SUBSCRIBE: Message.MessageType.ValueType  # 100
    DISCOVER_SUBSCRIBE_RESPONSE: Message.MessageType.ValueType  # 101

    class _ResponseStatus:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ResponseStatusEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Message._ResponseStatus.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        OK: Message._ResponseStatus.ValueType  # 0
        E_INVALID_NAMESPACE: Message._ResponseStatus.ValueType  # 100
        E_INVALID_PEER_INFO: Message._ResponseStatus.ValueType  # 101
        E_INVALID_TTL: Message._ResponseStatus.ValueType  # 102
        E_INVALID_COOKIE: Message._ResponseStatus.ValueType  # 103
        E_NOT_AUTHORIZED: Message._ResponseStatus.ValueType  # 200
        E_INTERNAL_ERROR: Message._ResponseStatus.ValueType  # 300
        E_UNAVAILABLE: Message._ResponseStatus.ValueType  # 400

    class ResponseStatus(_ResponseStatus, metaclass=_ResponseStatusEnumTypeWrapper): ...
    OK: Message.ResponseStatus.ValueType  # 0
    E_INVALID_NAMESPACE: Message.ResponseStatus.ValueType  # 100
    E_INVALID_PEER_INFO: Message.ResponseStatus.ValueType  # 101
    E_INVALID_TTL: Message.ResponseStatus.ValueType  # 102
    E_INVALID_COOKIE: Message.ResponseStatus.ValueType  # 103
    E_NOT_AUTHORIZED: Message.ResponseStatus.ValueType  # 200
    E_INTERNAL_ERROR: Message.ResponseStatus.ValueType  # 300
    E_UNAVAILABLE: Message.ResponseStatus.ValueType  # 400

    @typing.final
    class PeerInfo(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ID_FIELD_NUMBER: builtins.int
        ADDRS_FIELD_NUMBER: builtins.int
        id: builtins.bytes
        @property
        def addrs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
        def __init__(
            self,
            *,
            id: builtins.bytes = ...,
            addrs: collections.abc.Iterable[builtins.bytes] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["addrs", b"addrs", "id", b"id"]) -> None: ...

    @typing.final
    class Register(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NS_FIELD_NUMBER: builtins.int
        PEER_FIELD_NUMBER: builtins.int
        TTL_FIELD_NUMBER: builtins.int
        ns: builtins.str
        ttl: builtins.int
        """in seconds"""
        @property
        def peer(self) -> global___Message.PeerInfo: ...
        def __init__(
            self,
            *,
            ns: builtins.str = ...,
            peer: global___Message.PeerInfo | None = ...,
            ttl: builtins.int = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["peer", b"peer"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["ns", b"ns", "peer", b"peer", "ttl", b"ttl"]) -> None: ...

    @typing.final
    class RegisterResponse(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        STATUS_FIELD_NUMBER: builtins.int
        STATUSTEXT_FIELD_NUMBER: builtins.int
        TTL_FIELD_NUMBER: builtins.int
        status: global___Message.ResponseStatus.ValueType
        statusText: builtins.str
        ttl: builtins.int
        def __init__(
            self,
            *,
            status: global___Message.ResponseStatus.ValueType = ...,
            statusText: builtins.str = ...,
            ttl: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["status", b"status", "statusText", b"statusText", "ttl", b"ttl"]) -> None: ...

    @typing.final
    class Unregister(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NS_FIELD_NUMBER: builtins.int
        ID_FIELD_NUMBER: builtins.int
        ns: builtins.str
        id: builtins.bytes
        def __init__(
            self,
            *,
            ns: builtins.str = ...,
            id: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["id", b"id", "ns", b"ns"]) -> None: ...

    @typing.final
    class Discover(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NS_FIELD_NUMBER: builtins.int
        LIMIT_FIELD_NUMBER: builtins.int
        COOKIE_FIELD_NUMBER: builtins.int
        ns: builtins.str
        limit: builtins.int
        cookie: builtins.bytes
        def __init__(
            self,
            *,
            ns: builtins.str = ...,
            limit: builtins.int = ...,
            cookie: builtins.bytes = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["cookie", b"cookie", "limit", b"limit", "ns", b"ns"]) -> None: ...

    @typing.final
    class DiscoverResponse(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        REGISTRATIONS_FIELD_NUMBER: builtins.int
        COOKIE_FIELD_NUMBER: builtins.int
        STATUS_FIELD_NUMBER: builtins.int
        STATUSTEXT_FIELD_NUMBER: builtins.int
        cookie: builtins.bytes
        status: global___Message.ResponseStatus.ValueType
        statusText: builtins.str
        @property
        def registrations(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Message.Register]: ...
        def __init__(
            self,
            *,
            registrations: collections.abc.Iterable[global___Message.Register] | None = ...,
            cookie: builtins.bytes = ...,
            status: global___Message.ResponseStatus.ValueType = ...,
            statusText: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["cookie", b"cookie", "registrations", b"registrations", "status", b"status", "statusText", b"statusText"]) -> None: ...

    @typing.final
    class DiscoverSubscribe(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SUPPORTED_SUBSCRIPTION_TYPES_FIELD_NUMBER: builtins.int
        NS_FIELD_NUMBER: builtins.int
        ns: builtins.str
        @property
        def supported_subscription_types(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
        def __init__(
            self,
            *,
            supported_subscription_types: collections.abc.Iterable[builtins.str] | None = ...,
            ns: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["ns", b"ns", "supported_subscription_types", b"supported_subscription_types"]) -> None: ...

    @typing.final
    class DiscoverSubscribeResponse(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SUBSCRIPTION_TYPE_FIELD_NUMBER: builtins.int
        SUBSCRIPTION_DETAILS_FIELD_NUMBER: builtins.int
        STATUS_FIELD_NUMBER: builtins.int
        STATUSTEXT_FIELD_NUMBER: builtins.int
        subscription_type: builtins.str
        subscription_details: builtins.str
        status: global___Message.ResponseStatus.ValueType
        statusText: builtins.str
        def __init__(
            self,
            *,
            subscription_type: builtins.str = ...,
            subscription_details: builtins.str = ...,
            status: global___Message.ResponseStatus.ValueType = ...,
            statusText: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["status", b"status", "statusText", b"statusText", "subscription_details", b"subscription_details", "subscription_type", b"subscription_type"]) -> None: ...

    TYPE_FIELD_NUMBER: builtins.int
    REGISTER_FIELD_NUMBER: builtins.int
    REGISTERRESPONSE_FIELD_NUMBER: builtins.int
    UNREGISTER_FIELD_NUMBER: builtins.int
    DISCOVER_FIELD_NUMBER: builtins.int
    DISCOVERRESPONSE_FIELD_NUMBER: builtins.int
    DISCOVERSUBSCRIBE_FIELD_NUMBER: builtins.int
    DISCOVERSUBSCRIBERESPONSE_FIELD_NUMBER: builtins.int
    type: global___Message.MessageType.ValueType
    @property
    def register(self) -> global___Message.Register: ...
    @property
    def registerResponse(self) -> global___Message.RegisterResponse: ...
    @property
    def unregister(self) -> global___Message.Unregister: ...
    @property
    def discover(self) -> global___Message.Discover: ...
    @property
    def discoverResponse(self) -> global___Message.DiscoverResponse: ...
    @property
    def discoverSubscribe(self) -> global___Message.DiscoverSubscribe: ...
    @property
    def discoverSubscribeResponse(self) -> global___Message.DiscoverSubscribeResponse: ...
    def __init__(
        self,
        *,
        type: global___Message.MessageType.ValueType = ...,
        register: global___Message.Register | None = ...,
        registerResponse: global___Message.RegisterResponse | None = ...,
        unregister: global___Message.Unregister | None = ...,
        discover: global___Message.Discover | None = ...,
        discoverResponse: global___Message.DiscoverResponse | None = ...,
        discoverSubscribe: global___Message.DiscoverSubscribe | None = ...,
        discoverSubscribeResponse: global___Message.DiscoverSubscribeResponse | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["discover", b"discover", "discoverResponse", b"discoverResponse", "discoverSubscribe", b"discoverSubscribe", "discoverSubscribeResponse", b"discoverSubscribeResponse", "register", b"register", "registerResponse", b"registerResponse", "unregister", b"unregister"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["discover", b"discover", "discoverResponse", b"discoverResponse", "discoverSubscribe", b"discoverSubscribe", "discoverSubscribeResponse", b"discoverSubscribeResponse", "register", b"register", "registerResponse", b"registerResponse", "type", b"type", "unregister", b"unregister"]) -> None: ...

global___Message = Message

@typing.final
class RegistrationRecord(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    ADDRS_FIELD_NUMBER: builtins.int
    NS_FIELD_NUMBER: builtins.int
    TTL_FIELD_NUMBER: builtins.int
    id: builtins.str
    ns: builtins.str
    ttl: builtins.int
    @property
    def addrs(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        id: builtins.str = ...,
        addrs: collections.abc.Iterable[builtins.bytes] | None = ...,
        ns: builtins.str = ...,
        ttl: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["addrs", b"addrs", "id", b"id", "ns", b"ns", "ttl", b"ttl"]) -> None: ...

global___RegistrationRecord = RegistrationRecord



from collections.abc import Sequence
import logging

import trio

from nanolib.abc import IMultiselectClient, IMultiselectCommunicator
from nanolib.custom_types import TProtocol

from .exceptions import (
    MultiselectClientError,
    MultiselectCommunicatorError,
    ProtocolNotSupportedError,
)

logger = logging.getLogger("nanolib.protocol_muxer.multiselect_client")
logger.setLevel(logging.DEBUG)

MULTISELECT_PROTOCOL_ID = "/multistream/1.0.0"
PROTOCOL_NOT_FOUND_MSG = "na"
DEFAULT_NEGOTIATE_TIMEOUT = 30  # Increased for high-concurrency scenarios

logger = logging.getLogger(__name__)


class MultiselectClient(IMultiselectClient):
    """
    Client for communicating with receiver's multiselect module in order to
    select a protocol id to communicate over.
    """

    async def handshake(self, communicator: IMultiselectCommunicator) -> None:
        """
        Ensure that the client and multiselect are both using the same
        multiselect protocol.

        :param communicator: communicator to use to communicate with counterparty
        :raise MultiselectClientError: raised when handshake failed
        """
        try:
            await communicator.write(MULTISELECT_PROTOCOL_ID)
        except MultiselectCommunicatorError as error:
            raise MultiselectClientError(f"handshake write failed: {error}") from error

        try:
            handshake_contents = await communicator.read()

        except MultiselectCommunicatorError as error:
            raise MultiselectClientError(f"handshake read failed: {error}") from error

        if not is_valid_handshake(handshake_contents):
            raise MultiselectClientError(
                f"multiselect protocol ID mismatch: "
                f"expected {MULTISELECT_PROTOCOL_ID}, "
                f"got {handshake_contents!r}"
            )

    async def select_one_of(
        self,
        protocols: Sequence[TProtocol],
        communicator: IMultiselectCommunicator,
        negotiate_timeout: int = DEFAULT_NEGOTIATE_TIMEOUT,
    ) -> TProtocol:
        """
        For each protocol, send message to multiselect selecting protocol and
        fail if multiselect does not return same protocol. Returns first
        protocol that multiselect agrees on (i.e. that multiselect selects)

        :param protocols: protocols to select from
        :param communicator: communicator to use to communicate with counterparty
        :param negotiate_timeout: timeout for negotiation
        :return: selected protocol
        :raise MultiselectClientError: raised when protocol negotiation failed
        """
        try:
            with trio.fail_after(negotiate_timeout):
                await self.handshake(communicator)

                protocol_list = [str(p) for p in protocols]
                logger.debug(f"Attempting to negotiate one of: {protocol_list}")

                unsupported_errors: list[str] = []
                for protocol in protocols:
                    try:
                        selected_protocol = await self.try_select(
                            communicator, protocol
                        )
                        return selected_protocol
                    except ProtocolNotSupportedError as error:
                        unsupported_errors.append(str(error))
                        continue

                raise MultiselectClientError(
                    _build_protocols_not_supported_message(
                        protocols, negotiate_timeout, unsupported_errors
                    )
                )
        except trio.TooSlowError:
            raise MultiselectClientError(
                f"response timed out after {negotiate_timeout}s, "
                f"protocols tried: {list(protocols)}"
            )

    async def query_multistream_command(
        self,
        communicator: IMultiselectCommunicator,
        command: str,
        response_timeout: int = DEFAULT_NEGOTIATE_TIMEOUT,
    ) -> list[str]:
        """
        Send a multistream-select command over the given communicator and return
        parsed response.

        :param communicator: communicator to use to communicate with counterparty
        :param command: supported multistream-select command(e.g., ls)
        :param negotiate_timeout: timeout for negotiation
        :raise MultiselectClientError: If the communicator fails to process data.
        :return: list of strings representing the response from peer.
        """
        try:
            with trio.fail_after(response_timeout):
                await self.handshake(communicator)

                if command == "ls":
                    try:
                        await communicator.write("ls")
                    except MultiselectCommunicatorError as error:
                        raise MultiselectClientError(
                            f"command write failed: {error}, command={command}"
                        ) from error
                else:
                    raise ValueError("Command not supported")

                try:
                    response = await communicator.read()
                    response_list = response.strip().splitlines()

                except MultiselectCommunicatorError as error:
                    raise MultiselectClientError(
                        f"command read failed: {error}, command={command}"
                    ) from error

                return response_list
        except trio.TooSlowError:
            raise MultiselectClientError(
                f"command response timed out after {response_timeout}s, "
                f"command={command}"
            )

    async def try_select(
        self, communicator: IMultiselectCommunicator, protocol: TProtocol
    ) -> TProtocol:
        """
        Try to select the given protocol or raise exception if fails.

        :param communicator: communicator to use to communicate with counterparty
        :param protocol: protocol to select
        :raise MultiselectClientError: raised when protocol negotiation failed
        :return: selected protocol
        """
        # Represent `None` protocol as an empty string.
        protocol_str = protocol if protocol is not None else ""
        restart_attempted = False
        attempt = 0

        while True:
            attempt += 1
            try:
                await communicator.write(protocol_str)
            except MultiselectCommunicatorError as error:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "multiselect write failed for protocol %s on attempt %d: %s",
                        protocol,
                        attempt,
                        error,
                    )
                raise MultiselectClientError(
                    f"protocol write failed: {error}, protocol={protocol}"
                ) from error

            try:
                response = await communicator.read()
            except MultiselectCommunicatorError as error:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "multiselect read failed for protocol %s on attempt %d: %s",
                        protocol,
                        attempt,
                        error,
                    )
                raise MultiselectClientError(
                    f"protocol read failed: {error}, protocol={protocol}"
                ) from error

            if response == protocol_str:
                return protocol
            if response == PROTOCOL_NOT_FOUND_MSG:
                raise ProtocolNotSupportedError(
                    f"protocol not supported: {protocol}, response={response!r}"
                )
            if response == MULTISELECT_PROTOCOL_ID and not restart_attempted:
                # Some peers (notably go-libp2p during heavy churn) restart the
                # multistream handshake on existing streams. Re-run the handshake
                # once to resynchronize and retry the selection.
                restart_attempted = True
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(
                        "Peer restarted multiselect handshake mid-selection for %s "
                        "on attempt %d; re-running handshake",
                        protocol,
                        attempt,
                    )
                await self.handshake(communicator)
                continue
            raise MultiselectClientError(
                f"unrecognized response: {response!r}, expected {protocol_str!r} "
                f"or {PROTOCOL_NOT_FOUND_MSG!r}, protocol={protocol}"
            )
        raise MultiselectClientError(
            f"failed to negotiate protocol {protocol}: unexpected loop exit"
        )


def is_valid_handshake(handshake_contents: str) -> bool:
    """
    Determine if handshake is valid and should be confirmed.

    :param handshake_contents: contents of handshake message
    :return: true if handshake is complete, false otherwise
    """
    return handshake_contents == MULTISELECT_PROTOCOL_ID


def _build_protocols_not_supported_message(
    protocols: Sequence[TProtocol],
    negotiate_timeout: int,
    unsupported_errors: Sequence[str],
) -> str:
    """
    Format the final error message when every protocol was explicitly rejected.

    Includes the most recent rejection message to aid debugging so callers can
    see whether the peer closed the stream or responded with "na".
    """
    base = (
        f"protocols not supported: tried {list(protocols)}, "
        f"timeout={negotiate_timeout}s"
    )
    if unsupported_errors:
        return f"{base}. Last error: {unsupported_errors[-1]}"
    return base



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class KeyType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    RSA: _ClassVar[KeyType]
    Ed25519: _ClassVar[KeyType]
    Secp256k1: _ClassVar[KeyType]
    ECDSA: _ClassVar[KeyType]
RSA: KeyType
Ed25519: KeyType
Secp256k1: KeyType
ECDSA: KeyType

class PublicKey(_message.Message):
    __slots__ = ("Type", "Data")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    Type: KeyType
    Data: bytes
    def __init__(self, Type: _Optional[_Union[KeyType, str]] = ..., Data: _Optional[bytes] = ...) -> None: ...

class PrivateKey(_message.Message):
    __slots__ = ("Type", "Data")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    Type: KeyType
    Data: bytes
    def __init__(self, Type: _Optional[_Union[KeyType, str]] = ..., Data: _Optional[bytes] = ...) -> None: ...



"""
TLS security transport for libp2p.

This module provides a comprehensive TLS transport implementation
that follows the Go libp2p TLS specification.
"""

from nanolib.security.tls.transport import (
    TLSTransport,
    IdentityConfig,
    create_tls_transport,
    PROTOCOL_ID,
)
from nanolib.security.tls.io import TLSReadWriter
from nanolib.security.tls.certificate import (
    generate_certificate,
    create_cert_template,
    verify_certificate_chain,
    pub_key_from_cert_chain,
    SignedKey,
    ALPN_PROTOCOL
)

__all__ = [
    "TLSTransport",
    "IdentityConfig",
    "TLSReadWriter",
    "create_tls_transport",
    "generate_certificate",
    "create_cert_template",
    "verify_certificate_chain",
    "pub_key_from_cert_chain",
    "SignedKey",
    "PROTOCOL_ID",
    "ALPN_PROTOCOL"
]



"""
File chunking utilities for Merkle DAG.

This module provides functions to split files into chunks for storage
in a Merkle DAG structure, enabling efficient storage and transfer of
large files.
"""

from collections.abc import Callable, Iterator
from pathlib import Path

# Default chunk size: 63 KB (py-libp2p accepts less than 64 KB)
DEFAULT_CHUNK_SIZE = 63 * 1024


def chunk_bytes(data: bytes, chunk_size: int = DEFAULT_CHUNK_SIZE) -> list[bytes]:
    """
    Split bytes into fixed-size chunks.

    Args:
        data: Data to chunk
        chunk_size: Size of each chunk in bytes

    Returns:
        List of chunks

    Example:
        >>> data = b"x" * 1_000_000  # 1 MB
        >>> chunks = chunk_bytes(data, chunk_size=256*1024)
        >>> print(f"Split into {len(chunks)} chunks")
        Split into 4 chunks
        >>> print(f"Last chunk size: {len(chunks[-1])}")
        Last chunk size: 230400

    """
    if not data:
        return []

    chunks = []
    offset = 0

    while offset < len(data):
        chunk = data[offset : offset + chunk_size]
        chunks.append(chunk)
        offset += chunk_size

    return chunks


def chunk_file(file_path: str, chunk_size: int = DEFAULT_CHUNK_SIZE) -> Iterator[bytes]:
    """
    Stream file in chunks (memory efficient).

    This is the recommended way to chunk large files as it doesn't
    load the entire file into memory.

    Args:
        file_path: Path to file
        chunk_size: Size of each chunk in bytes

    Yields:
        File chunks

    Raises:
        FileNotFoundError: If file doesn't exist

    Example:
        >>> for i, chunk in enumerate(chunk_file('large_file.bin')):
        ...     print(f"Processing chunk {i}: {len(chunk)} bytes")
        ...     # Process chunk

    """
    path = Path(file_path)
    if not path.exists():
        raise FileNotFoundError(f"File not found: {file_path}")

    with open(file_path, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            yield chunk


def estimate_chunk_count(file_size: int, chunk_size: int = DEFAULT_CHUNK_SIZE) -> int:
    """
    Estimate number of chunks for a given file size.

    Args:
        file_size: Size of file in bytes
        chunk_size: Chunk size in bytes (defaults to DEFAULT_CHUNK_SIZE)

    Returns:
        Estimated number of chunks

    Example:
        >>> estimate_chunk_count(1024 * 1024)  # 1 MB file
        17
        >>> estimate_chunk_count(500 * 1024 * 1024)  # 500 MB file
        8127

    """
    if file_size < 0:
        raise ValueError(f"file_size must be non-negative, got {file_size}")

    if file_size == 0:
        return 0

    # Round up division
    return (file_size + chunk_size - 1) // chunk_size


def get_file_size(file_path: str) -> int:
    """
    Get the size of a file.

    Args:
        file_path: Path to file

    Returns:
        File size in bytes

    Raises:
        FileNotFoundError: If file doesn't exist

    """
    path = Path(file_path)
    if not path.exists():
        raise FileNotFoundError(f"File not found: {file_path}")
    return path.stat().st_size


def chunk_file_with_progress(
    file_path: str,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    progress_callback: Callable[[int, int, int], None] | None = None,
) -> Iterator[bytes]:
    """
    Stream file in chunks with progress tracking.

    Args:
        file_path: Path to file
        chunk_size: Size of each chunk in bytes
        progress_callback: Optional callback(current_byte, total_bytes, chunk_num)

    Yields:
        File chunks

    Example:
        >>> def on_progress(current, total, chunk_num):
        ...     percent = (current / total) * 100
        ...     print(f"Progress: {percent:.1f}% (chunk {chunk_num})")
        >>> for chunk in chunk_file_with_progress(
        ...     'file.bin', progress_callback=on_progress
        ... ):
        ...     process(chunk)

    """
    file_size = get_file_size(file_path)
    bytes_read = 0
    chunk_num = 0

    for chunk in chunk_file(file_path, chunk_size):
        bytes_read += len(chunk)
        chunk_num += 1

        if progress_callback:
            progress_callback(bytes_read, file_size, chunk_num)

        yield chunk


__all__ = [
    "DEFAULT_CHUNK_SIZE",
    "chunk_bytes",
    "chunk_file",
    "chunk_file_with_progress",
    "estimate_chunk_count",
    "get_file_size",
]



"""
Optimized resource metrics implementation.

This module provides high-performance metrics collection for the resource manager,
using array-based storage for optimal performance in production environments.
"""

from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from enum import IntEnum
import threading
import time
from typing import Any


class Direction(IntEnum):
    """Direction enum for resource tracking"""

    INBOUND = 0
    OUTBOUND = 1

    def __str__(self) -> str:
        """Return string representation as numeric value."""
        return str(self.value)


class MetricType(IntEnum):
    """Enumeration of metric types for array indexing."""

    CONNECTIONS_INBOUND = 0
    CONNECTIONS_OUTBOUND = 1
    MEMORY_USAGE = 2
    STREAMS_INBOUND = 3
    STREAMS_OUTBOUND = 4
    CONNECTION_BLOCKS = 5
    MEMORY_BLOCKS = 6
    STREAM_BLOCKS = 7
    PEAK_CONNECTIONS = 8
    PEAK_MEMORY = 9
    PEAK_STREAMS = 10

    def __str__(self) -> str:
        """Return string representation as numeric value."""
        return str(self.value)


@dataclass
class ResourceMetrics:
    """Metrics for a specific resource type."""

    allowed: int = 0
    blocked: int = 0
    current_usage: int = 0
    peak_usage: int = 0
    last_updated: float = field(default_factory=time.time)


class Metrics:
    """
    High-performance metrics with array storage.

    This implementation uses pre-allocated arrays for O(1) access
    to metrics, avoiding dictionary overhead and string operations.
    """

    def __init__(self) -> None:
        """Initialize the optimized metrics collector."""
        # Pre-allocated arrays for O(1) access
        self._counters = [0] * len(MetricType)
        self._gauges = [0.0] * len(MetricType)
        self._lock = threading.RLock()
        self._start_time = time.time()

        # Resource-specific metrics (for backward compatibility)
        self.resource_metrics: dict[str, ResourceMetrics] = defaultdict(ResourceMetrics)

    def increment(self, metric: MetricType, delta: int = 1) -> None:
        """
        Increment a counter metric.

        Args:
            metric: The metric type to increment
            delta: Amount to increment by

        """
        with self._lock:
            self._counters[metric] += delta

    def set_gauge(self, metric: MetricType, value: float) -> None:
        """
        Set a gauge metric value.

        Args:
            metric: The metric type to set
            value: The value to set

        """
        with self._lock:
            self._gauges[metric] = value

    def get_counter(self, metric: MetricType) -> int:
        """
        Get a counter metric value.

        Args:
            metric: The metric type to get

        Returns:
            The counter value

        """
        with self._lock:
            return self._counters[metric]

    def get_gauge(self, metric: MetricType) -> float:
        """
        Get a gauge metric value.

        Args:
            metric: The metric type to get

        Returns:
            The gauge value

        """
        with self._lock:
            return self._gauges[metric]

    def record_connection(self, direction: str, delta: int = 1) -> None:
        """
        Record a connection event.

        Args:
            direction: Connection direction ('inbound' or 'outbound')
            delta: Change in connection count

        """
        if direction == "inbound":
            self.increment(MetricType.CONNECTIONS_INBOUND, delta)
        else:
            self.increment(MetricType.CONNECTIONS_OUTBOUND, delta)

        # Update peak connections
        current_total = self.get_counter(
            MetricType.CONNECTIONS_INBOUND
        ) + self.get_counter(MetricType.CONNECTIONS_OUTBOUND)
        if current_total > self.get_gauge(MetricType.PEAK_CONNECTIONS):
            self.set_gauge(MetricType.PEAK_CONNECTIONS, current_total)

    def record_memory(self, size: int, delta: int = 1) -> None:
        """
        Record a memory event.

        Args:
            size: Memory size in bytes
            delta: Change in memory count

        """
        self.increment(MetricType.MEMORY_USAGE, size * delta)

        # Update peak memory
        current_memory = self.get_counter(MetricType.MEMORY_USAGE)
        if current_memory > self.get_gauge(MetricType.PEAK_MEMORY):
            self.set_gauge(MetricType.PEAK_MEMORY, current_memory)

    def record_stream(self, direction: str, delta: int = 1) -> None:
        """
        Record a stream event.

        Args:
            direction: Stream direction ('inbound' or 'outbound')
            delta: Change in stream count

        """
        if direction == "inbound":
            self.increment(MetricType.STREAMS_INBOUND, delta)
        else:
            self.increment(MetricType.STREAMS_OUTBOUND, delta)

        # Update peak streams
        current_total = self.get_counter(MetricType.STREAMS_INBOUND) + self.get_counter(
            MetricType.STREAMS_OUTBOUND
        )
        if current_total > self.get_gauge(MetricType.PEAK_STREAMS):
            self.set_gauge(MetricType.PEAK_STREAMS, current_total)

    def record_block(self, resource_type: str) -> None:
        """
        Record a resource block event.

        Args:
            resource_type: Type of resource that was blocked

        """
        if resource_type == "connection":
            self.increment(MetricType.CONNECTION_BLOCKS)
        elif resource_type == "memory":
            self.increment(MetricType.MEMORY_BLOCKS)
        elif resource_type == "stream":
            self.increment(MetricType.STREAM_BLOCKS)

    # Backward compatibility methods
    def allow_conn(self, direction: str, use_fd: bool = True) -> None:
        """Record an allowed connection (backward compatibility)."""
        self.record_connection(direction)

    def remove_conn(self, direction: str, use_fd: bool = True) -> None:
        """Record a removed connection (backward compatibility)."""
        self.record_connection(direction, -1)

    def allow_stream(self, peer_id: str, direction: str) -> None:
        """Record an allowed stream (backward compatibility)."""
        self.record_stream(direction)

    def remove_stream(self, direction: str) -> None:
        """Record a removed stream (backward compatibility)."""
        self.record_stream(direction, -1)

    def allow_memory(self, size: int) -> None:
        """Record allowed memory (backward compatibility)."""
        self.record_memory(size)

    def release_memory(self, size: int) -> None:
        """Record released memory (backward compatibility)."""
        self.record_memory(size, -1)

    def block_conn(self, direction: str, use_fd: bool = True) -> None:
        """Record a blocked connection (backward compatibility)."""
        self.record_block("connection")

    def block_stream(self, peer_id: str, direction: str) -> None:
        """Record a blocked stream (backward compatibility)."""
        self.record_block("stream")

    def block_memory(self, size: int) -> None:
        """Record blocked memory (backward compatibility)."""
        self.record_block("memory")

    def get_summary(self) -> dict[str, Any]:
        """
        Get a summary of all metrics.

        Returns:
            Dictionary containing all metric values

        """
        with self._lock:
            return {
                "uptime": time.time() - self._start_time,
                "connections": {
                    "inbound": self.get_counter(MetricType.CONNECTIONS_INBOUND),
                    "outbound": self.get_counter(MetricType.CONNECTIONS_OUTBOUND),
                    "total": (
                        self.get_counter(MetricType.CONNECTIONS_INBOUND)
                        + self.get_counter(MetricType.CONNECTIONS_OUTBOUND)
                    ),
                    "peak": self.get_gauge(MetricType.PEAK_CONNECTIONS),
                },
                "memory": {
                    "current": self.get_counter(MetricType.MEMORY_USAGE),
                    "peak": self.get_gauge(MetricType.PEAK_MEMORY),
                },
                "streams": {
                    "inbound": self.get_counter(MetricType.STREAMS_INBOUND),
                    "outbound": self.get_counter(MetricType.STREAMS_OUTBOUND),
                    "total": (
                        self.get_counter(MetricType.STREAMS_INBOUND)
                        + self.get_counter(MetricType.STREAMS_OUTBOUND)
                    ),
                    "peak": self.get_gauge(MetricType.PEAK_STREAMS),
                },
                "blocks": {
                    "connections": self.get_counter(MetricType.CONNECTION_BLOCKS),
                    "memory": self.get_counter(MetricType.MEMORY_BLOCKS),
                    "streams": self.get_counter(MetricType.STREAM_BLOCKS),
                },
            }

    def reset(self) -> None:
        """Reset all metrics to zero."""
        with self._lock:
            self._counters = [0] * len(MetricType)
            self._gauges = [0.0] * len(MetricType)
        self.resource_metrics.clear()
        self._start_time = time.time()



from typing import Any

from .tcp.tcp import TCP
from .websocket.transport import WebsocketTransport
from .transport_registry import (
    TransportRegistry,
    create_transport_for_multiaddr,
    get_transport_registry,
    register_transport,
    get_supported_transport_protocols,
)
from .upgrader import TransportUpgrader
from nanolib.abc import ITransport

def create_transport(protocol: str, upgrader: TransportUpgrader | None = None, **kwargs: Any) -> ITransport:
    """
    Convenience function to create a transport instance.

    :param protocol: The transport protocol ("tcp", "ws", "wss", or custom)
    :param upgrader: Optional transport upgrader (required for WebSocket)
    :param kwargs: Additional arguments for transport construction (e.g., tls_client_config, tls_server_config)
    :return: Transport instance
    """
    # First check if it's a built-in protocol
    if protocol in ["ws", "wss"]:
        if upgrader is None:
            raise ValueError(f"WebSocket transport requires an upgrader")
        from nanolib.transport.websocket.transport import WebsocketConfig, WebsocketTransport
        config = WebsocketConfig(
            tls_client_config=kwargs.get("tls_client_config"),
            tls_server_config=kwargs.get("tls_server_config"),
            handshake_timeout=kwargs.get("handshake_timeout", 15.0)
        )
        return WebsocketTransport(upgrader, config)
    elif protocol == "tcp":
        return TCP()
    else:
        # Check if it's a custom registered transport
        registry = get_transport_registry()
        transport_class = registry.get_transport(protocol)
        if transport_class:
            transport = registry.create_transport(protocol, upgrader, **kwargs)
            if transport is None:
                raise ValueError(f"Failed to create transport for protocol: {protocol}")
            return transport
        else:
            raise ValueError(f"Unsupported transport protocol: {protocol}")

__all__ = [
    "TCP",
    "WebsocketTransport",
    "TransportRegistry",
    "create_transport_for_multiaddr",
    "create_transport",
    "get_transport_registry",
    "register_transport",
    "get_supported_transport_protocols",
]



"""
Enhanced Noise transport implementation.

This module provides an enhanced Noise transport with support for advanced
features including early data, WebTransport integration, and rekey management.
The transport uses the XX handshake pattern for mutual authentication and
forward secrecy.
"""

from nanolib.abc import (
    IRawConnection,
    ISecureConn,
    ISecureTransport,
)
from nanolib.crypto.keys import (
    KeyPair,
    PrivateKey,
)
from nanolib.custom_types import (
    TProtocol,
)
from nanolib.peer.id import (
    ID,
)

from .early_data import EarlyDataHandler, EarlyDataManager
from .patterns import (
    IPattern,
    PatternXX,
)
from .rekey import RekeyManager, RekeyPolicy
from .webtransport import WebTransportSupport

PROTOCOL_ID = TProtocol("/noise")


class Transport(ISecureTransport):
    """Enhanced Noise transport with advanced features support."""

    libp2p_privkey: PrivateKey
    noise_privkey: PrivateKey
    local_peer: ID
    early_data: bytes | None
    webtransport_support: WebTransportSupport
    early_data_manager: EarlyDataManager
    rekey_manager: RekeyManager

    def __init__(
        self,
        libp2p_keypair: KeyPair,
        noise_privkey: PrivateKey,
        early_data: bytes | None = None,
        early_data_handler: EarlyDataHandler | None = None,
        rekey_policy: RekeyPolicy | None = None,
    ) -> None:
        """
        Initialize enhanced Noise transport.

        Args:
            libp2p_keypair: libp2p key pair
            noise_privkey: Noise private key
            early_data: Optional early data
            early_data_handler: Optional early data handler
            rekey_policy: Optional rekey policy

        """
        self.libp2p_privkey = libp2p_keypair.private_key
        self.noise_privkey = noise_privkey
        self.local_peer = ID.from_pubkey(libp2p_keypair.public_key)
        self.early_data = early_data

        # Initialize advanced features
        self.webtransport_support = WebTransportSupport()
        self.early_data_manager = EarlyDataManager(early_data_handler)
        self.rekey_manager = RekeyManager(rekey_policy)
        self._static_key_cache: dict[ID, bytes] = {}

    def get_pattern(self) -> IPattern:
        """
        Get the handshake pattern for the connection.

        Returns:
            IPattern: The XX handshake pattern

        """
        # Always use XX pattern (IK pattern has been deprecated)
        return PatternXX(
            self.local_peer,
            self.libp2p_privkey,
            self.noise_privkey,
            self.early_data,
        )

    async def secure_inbound(self, conn: IRawConnection) -> ISecureConn:
        """
        Perform inbound secure connection.

        Args:
            conn: Raw connection

        Returns:
            ISecureConn: Secure connection

        """
        pattern = self.get_pattern()
        secure_conn = await pattern.handshake_inbound(conn)

        # Handle early data if present
        early_data = getattr(pattern, "early_data", None)
        if early_data is not None:
            await self.early_data_manager.handle_early_data(early_data)

        return secure_conn

    async def secure_outbound(self, conn: IRawConnection, peer_id: ID) -> ISecureConn:
        """
        Perform outbound secure connection.

        Args:
            conn: Raw connection
            peer_id: Remote peer ID

        Returns:
            ISecureConn: Secure connection

        """
        pattern = self.get_pattern()
        secure_conn = await pattern.handshake_outbound(conn, peer_id)

        # Handle early data if present
        early_data = getattr(pattern, "early_data", None)
        if early_data is not None:
            await self.early_data_manager.handle_early_data(early_data)

        return secure_conn

    def cache_static_key(self, peer_id: ID, static_key: bytes) -> None:
        """
        Cache a static key for a peer.

        Args:
            peer_id: The peer ID
            static_key: The static key to cache

        """
        self._static_key_cache[peer_id] = static_key

    def get_cached_static_key(self, peer_id: ID) -> bytes | None:
        """
        Get cached static key for a peer.

        Args:
            peer_id: The peer ID

        Returns:
            The cached static key or None if not found

        """
        return self._static_key_cache.get(peer_id)

    def clear_static_key_cache(self) -> None:
        """Clear the static key cache."""
        self._static_key_cache.clear()



from collections.abc import (
    Callable,
)

from nanolib.abc import (
    PeerInfo,
)

TTL: int = 60 * 60  # Time-to-live for discovered peers in seconds


class PeerDiscovery:
    def __init__(self) -> None:
        self._peer_discovered_handlers: list[Callable[[PeerInfo], None]] = []

    def register_peer_discovered_handler(
        self, handler: Callable[[PeerInfo], None]
    ) -> None:
        self._peer_discovered_handlers.append(handler)

    def emit_peer_discovered(self, peer_info: PeerInfo) -> None:
        for handler in self._peer_discovered_handlers:
            handler(peer_info)


peerDiscovery = PeerDiscovery()



"""
SQLite datastore implementation for persistent peer storage.

This provides a SQLite-based datastore for persistent peer storage.
"""

from collections.abc import Iterator
from pathlib import Path
import sqlite3

import trio

from .base import IBatch, IBatchingDatastore


class SQLiteBatch(IBatch):
    """SQLite batch implementation."""

    def __init__(self, connection: sqlite3.Connection):
        self.connection = connection
        self.operations: list[tuple[str, bytes, bytes | None]] = []

    async def put(self, key: bytes, value: bytes) -> None:
        """Add a put operation to the batch."""
        self.operations.append(("put", key, value))

    async def delete(self, key: bytes) -> None:
        """Add a delete operation to the batch."""
        self.operations.append(("delete", key, None))

    async def commit(self) -> None:
        """Commit all operations in the batch."""
        try:
            cursor = self.connection.cursor()
            for operation, key, value in self.operations:
                if operation == "put":
                    cursor.execute(
                        "INSERT OR REPLACE INTO datastore (key, value) VALUES (?, ?)",
                        (key, value),
                    )
                elif operation == "delete":
                    cursor.execute("DELETE FROM datastore WHERE key = ?", (key,))
            self.connection.commit()
        except Exception as e:
            self.connection.rollback()
            raise e

    async def discard(self) -> None:
        """Discard all operations in the batch."""
        self.operations.clear()


class SQLiteDatastore(IBatchingDatastore):
    """
    SQLite-based datastore implementation.

    This provides persistent storage using SQLite database files.
    Supports context manager protocol for proper resource management.
    """

    def __init__(self, path: str | Path):
        """
        Initialize SQLite datastore.

        Args:
            path: Path to the SQLite database file

        """
        self.path = Path(path)
        self.connection: sqlite3.Connection | None = None
        self._lock = trio.Lock()
        self._closed = False

    async def _ensure_connection(self) -> None:
        """
        Ensure database connection is established.

        :raises RuntimeError: If datastore is closed
        :raises sqlite3.Error: If database connection fails
        """
        if self._closed:
            raise RuntimeError("Datastore is closed")

        if self.connection is None:
            async with self._lock:
                if self.connection is None:
                    # Create directory if it doesn't exist
                    self.path.parent.mkdir(parents=True, exist_ok=True)

                    # Set appropriate file permissions (readable/writable by owner only)
                    if not self.path.exists():
                        self.path.touch(mode=0o600)

                    self.connection = sqlite3.connect(
                        str(self.path),
                        check_same_thread=False,
                        timeout=30.0,  # Add timeout for database locks
                    )

                    # Enable WAL mode for better concurrency
                    cursor = self.connection.cursor()
                    cursor.execute("PRAGMA journal_mode=WAL")
                    cursor.execute(
                        "PRAGMA synchronous=NORMAL"
                    )  # Balance between safety and performance
                    cursor.execute("PRAGMA cache_size=10000")  # Increase cache size
                    cursor.execute(
                        "PRAGMA temp_store=MEMORY"
                    )  # Use memory for temp storage

                    # Create table if it doesn't exist
                    cursor.execute("""
                        CREATE TABLE IF NOT EXISTS datastore (
                            key BLOB PRIMARY KEY,
                            value BLOB NOT NULL
                        )
                    """)
                    self.connection.commit()

    async def get(self, key: bytes) -> bytes | None:
        """Retrieve a value by key."""
        await self._ensure_connection()
        if self.connection is None:
            raise ValueError("SQLite connection is not initialized")
        cursor = self.connection.cursor()
        cursor.execute("SELECT value FROM datastore WHERE key = ?", (key,))
        result = cursor.fetchone()
        return result[0] if result else None

    async def put(self, key: bytes, value: bytes) -> None:
        """Store a key-value pair."""
        await self._ensure_connection()
        if self.connection is None:
            raise ValueError("SQLite connection is not initialized")
        cursor = self.connection.cursor()
        cursor.execute(
            "INSERT OR REPLACE INTO datastore (key, value) VALUES (?, ?)", (key, value)
        )
        self.connection.commit()

    async def delete(self, key: bytes) -> None:
        """Delete a key-value pair."""
        await self._ensure_connection()
        if self.connection is None:
            raise ValueError("SQLite connection is not initialized")
        cursor = self.connection.cursor()
        cursor.execute("DELETE FROM datastore WHERE key = ?", (key,))
        self.connection.commit()

    async def has(self, key: bytes) -> bool:
        """Check if a key exists."""
        await self._ensure_connection()
        if self.connection is None:
            raise ValueError("SQLite connection is not initialized")
        cursor = self.connection.cursor()
        cursor.execute("SELECT 1 FROM datastore WHERE key = ?", (key,))
        return cursor.fetchone() is not None

    def query(self, prefix: bytes = b"") -> Iterator[tuple[bytes, bytes]]:
        """
        Query key-value pairs with optional prefix.

        Note: query is synchronous per interface and returns an Iterator.
        If the connection is missing, we best-effort open it synchronously.
        """
        if self.connection is None:
            # Create directory and open connection synchronously
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self.connection = sqlite3.connect(str(self.path), check_same_thread=False)
            self.connection.execute(
                """
                        CREATE TABLE IF NOT EXISTS datastore (
                            key BLOB PRIMARY KEY,
                            value BLOB NOT NULL
                        )
                    """
            )
            self.connection.commit()

        if self.connection is None:
            raise ValueError("SQLite connection is not initialized")
        cursor = self.connection.cursor()
        if prefix:
            cursor.execute(
                "SELECT key, value FROM datastore WHERE key LIKE ?", (prefix + b"%",)
            )
        else:
            cursor.execute("SELECT key, value FROM datastore")

        for row in cursor:
            yield row[0], row[1]

    async def batch(self) -> IBatch:
        """Create a new batch for atomic operations."""
        await self._ensure_connection()
        if self.connection is None:
            raise ValueError("SQLite connection is not initialized")
        return SQLiteBatch(self.connection)

    async def sync(self, prefix: bytes) -> None:
        """Flush pending writes to disk (commit current transaction)."""
        await self._ensure_connection()
        if self.connection is None:
            raise ValueError("SQLite connection is not initialized")
        self.connection.commit()

    async def close(self) -> None:
        """
        Close the datastore connection.

        This method is idempotent and can be called multiple times safely.
        """
        async with self._lock:
            if self.connection and not self._closed:
                try:
                    self.connection.close()
                finally:
                    self.connection = None
                    self._closed = True

    async def __aenter__(self) -> "SQLiteDatastore":
        """Async context manager entry."""
        await self._ensure_connection()
        return self

    async def __aexit__(
        self, exc_type: type, exc_val: Exception, exc_tb: object
    ) -> None:
        """Async context manager exit."""
        await self.close()



"""WebTransport integration for Noise protocol extensions."""

import hashlib
from typing import Protocol, runtime_checkable


@runtime_checkable
class WebTransportCertificate(Protocol):
    """Protocol for WebTransport certificate handling."""

    def get_certificate_hash(self) -> bytes:
        """
        Get the SHA-256 hash of the certificate.

        Returns:
            bytes: The SHA-256 hash of the certificate

        """
        ...


class WebTransportCertManager:
    """Manager for WebTransport certificate hashes in Noise extensions."""

    def __init__(self) -> None:
        self._cert_hashes: list[bytes] = []

    def add_certificate(self, cert_data: bytes) -> bytes:
        """
        Add a certificate and return its hash.

        Args:
            cert_data: The certificate data

        Returns:
            bytes: The SHA-256 hash of the certificate

        """
        cert_hash = hashlib.sha256(cert_data).digest()
        if cert_hash not in self._cert_hashes:
            self._cert_hashes.append(cert_hash)
        return cert_hash

    def add_certificate_hash(self, cert_hash: bytes) -> None:
        """
        Add a certificate hash directly.

        Args:
            cert_hash: The SHA-256 hash of the certificate

        """
        if cert_hash not in self._cert_hashes:
            self._cert_hashes.append(cert_hash)

    def get_certificate_hashes(self) -> list[bytes]:
        """
        Get all certificate hashes.

        Returns:
            list[bytes]: List of certificate hashes

        """
        return self._cert_hashes.copy()

    def has_certificate_hash(self, cert_hash: bytes) -> bool:
        """
        Check if a certificate hash is present.

        Args:
            cert_hash: The certificate hash to check

        Returns:
            bool: True if the certificate hash is present

        """
        return cert_hash in self._cert_hashes

    def clear_certificates(self) -> None:
        """Clear all certificate hashes."""
        self._cert_hashes.clear()

    def __len__(self) -> int:
        """Get the number of certificate hashes."""
        return len(self._cert_hashes)

    def __bool__(self) -> bool:
        """Check if there are any certificate hashes."""
        return bool(self._cert_hashes)


class WebTransportSupport:
    """Support for WebTransport integration in Noise protocol."""

    def __init__(self) -> None:
        self._cert_manager = WebTransportCertManager()

    @property
    def cert_manager(self) -> WebTransportCertManager:
        """Get the certificate manager."""
        return self._cert_manager

    def add_certificate(self, cert_data: bytes) -> bytes:
        """
        Add a WebTransport certificate.

        Args:
            cert_data: The certificate data

        Returns:
            bytes: The SHA-256 hash of the certificate

        """
        return self._cert_manager.add_certificate(cert_data)

    def get_certificate_hashes(self) -> list[bytes]:
        """
        Get all certificate hashes for extensions.

        Returns:
            list[bytes]: List of certificate hashes

        """
        return self._cert_manager.get_certificate_hashes()

    def has_certificates(self) -> bool:
        """
        Check if any certificates are configured.

        Returns:
            bool: True if certificates are configured

        """
        return bool(self._cert_manager)

    def validate_certificate_hash(self, cert_hash: bytes) -> bool:
        """
        Validate a certificate hash.

        Args:
            cert_hash: The certificate hash to validate

        Returns:
            bool: True if the certificate hash is valid

        """
        # Basic validation: check if it's a valid SHA-256 hash (32 bytes)
        if len(cert_hash) != 32:
            return False

        # Check if it's a known certificate hash
        return self._cert_manager.has_certificate_hash(cert_hash)



from abc import (
    ABC,
    abstractmethod,
)
from typing import Any


class Closer(ABC):
    @abstractmethod
    async def close(self) -> None: ...


class Reader(ABC):
    @abstractmethod
    async def read(self, n: int | None = None) -> bytes: ...


class Writer(ABC):
    @abstractmethod
    async def write(self, data: bytes) -> None: ...


class WriteCloser(Writer, Closer):
    pass


class ReadCloser(Reader, Closer):
    pass


class ReadWriter(Reader, Writer):
    pass


class ReadWriteCloser(Reader, Writer, Closer):
    @abstractmethod
    def get_remote_address(self) -> tuple[str, int] | None:
        """
        Return the remote address of the connected peer.

        :return: A tuple of (host, port) or None if not available
        """
        ...


class MsgReader(ABC):
    @abstractmethod
    async def read_msg(self) -> bytes: ...


class MsgWriter(ABC):
    @abstractmethod
    async def write_msg(self, msg: bytes) -> None: ...


class MsgReadWriteCloser(MsgReader, MsgWriter, Closer):
    pass


class Encrypter(ABC):
    @abstractmethod
    def encrypt(self, data: bytes) -> bytes: ...

    @abstractmethod
    def decrypt(self, data: bytes) -> bytes: ...


class EncryptedMsgReadWriter(MsgReadWriteCloser, Encrypter):
    """Read/write message with encryption/decryption."""

    conn: Any | None

    def __init__(self, conn: Any | None = None):
        self.conn = conn

    def get_remote_address(self) -> tuple[str, int] | None:
        """Get remote address if supported by the underlying connection."""
        if (
            self.conn is not None
            and hasattr(self, "conn")
            and hasattr(self.conn, "get_remote_address")
        ):
            return self.conn.get_remote_address()
        return None



"""
Memory statistics caching for resource management.

This module implements memory statistics caching and monitoring to support
memory-based connection limits, matching the Rust libp2p-memory-connection-limits
behavior.
"""

from __future__ import annotations

from dataclasses import dataclass
import threading
import time

import psutil


@dataclass
class MemoryStats:
    """Memory statistics snapshot."""

    # Process memory
    process_memory_bytes: int
    process_memory_percent: float

    # System memory
    system_memory_total: int
    system_memory_available: int
    system_memory_percent: float

    # Timestamp
    timestamp: float

    @property
    def process_memory_mb(self) -> float:
        """Process memory in megabytes."""
        return self.process_memory_bytes / (1024 * 1024)

    @property
    def system_memory_total_gb(self) -> float:
        """System total memory in gigabytes."""
        return self.system_memory_total / (1024 * 1024 * 1024)

    @property
    def system_memory_available_gb(self) -> float:
        """System available memory in gigabytes."""
        return self.system_memory_available / (1024 * 1024 * 1024)


class MemoryStatsCache:
    """
    Cache and monitor memory statistics for resource management.

    This class provides memory statistics caching and monitoring to support
    memory-based connection limits, matching the Rust libp2p-memory-connection-limits
    behavior.
    """

    def __init__(self, cache_duration: float = 1.0):
        """
        Initialize memory stats cache.

        Args:
            cache_duration: Duration in seconds to cache memory stats

        """
        self.cache_duration = cache_duration
        self._lock = threading.RLock()
        self._cached_stats: MemoryStats | None = None
        self._last_update: float = 0.0
        self._cached_summary: dict[str, object] | None = None

        # Process object for monitoring
        try:
            self._process: psutil.Process = psutil.Process()
        except Exception:
            # If psutil fails, we'll handle it in get_memory_stats
            self._process = None  # type: ignore

    def get_memory_stats(self, force_refresh: bool = False) -> MemoryStats:
        """
        Get current memory statistics.

        Args:
            force_refresh: Force refresh of cached statistics

        Returns:
            MemoryStats: Current memory statistics

        """
        with self._lock:
            current_time = time.time()

            # Return cached stats if still valid
            if (
                not force_refresh
                and self._cached_stats is not None
                and (current_time - self._last_update) < self.cache_duration
            ):
                return self._cached_stats

            # Refresh memory statistics
            self._cached_stats = self._collect_memory_stats()
            self._last_update = current_time

            return self._cached_stats

    def _collect_memory_stats(self) -> MemoryStats:
        """
        Collect current memory statistics.

        Returns:
            MemoryStats: Current memory statistics

        """
        if self._process is None:
            # Return default values if psutil failed
            return MemoryStats(
                process_memory_bytes=0,
                process_memory_percent=0.0,
                system_memory_total=0,
                system_memory_available=0,
                system_memory_percent=0.0,
                timestamp=time.time(),
            )

        # Process memory
        process_memory_info = self._process.memory_info()  # type: ignore
        process_memory_bytes = process_memory_info.rss  # Resident Set Size

        # System memory
        system_memory = psutil.virtual_memory()

        return MemoryStats(
            process_memory_bytes=process_memory_bytes,
            process_memory_percent=self._process.memory_percent(),  # type: ignore
            system_memory_total=system_memory.total,
            system_memory_available=system_memory.available,
            system_memory_percent=system_memory.percent,
            timestamp=time.time(),
        )

    def get_process_memory_bytes(self, force_refresh: bool = False) -> int:
        """
        Get current process memory usage in bytes.

        Args:
            force_refresh: Force refresh of cached statistics

        Returns:
            int: Process memory usage in bytes

        """
        stats = self.get_memory_stats(force_refresh)
        return stats.process_memory_bytes

    def get_process_memory_percent(self, force_refresh: bool = False) -> float:
        """
        Get current process memory usage as percentage of system memory.

        Args:
            force_refresh: Force refresh of cached statistics

        Returns:
            float: Process memory usage percentage

        """
        stats = self.get_memory_stats(force_refresh)
        return stats.process_memory_percent

    def get_system_memory_total(self, force_refresh: bool = False) -> int:
        """
        Get total system memory in bytes.

        Args:
            force_refresh: Force refresh of cached statistics

        Returns:
            int: Total system memory in bytes

        """
        stats = self.get_memory_stats(force_refresh)
        return stats.system_memory_total

    def get_system_memory_available(self, force_refresh: bool = False) -> int:
        """
        Get available system memory in bytes.

        Args:
            force_refresh: Force refresh of cached statistics

        Returns:
            int: Available system memory in bytes

        """
        stats = self.get_memory_stats(force_refresh)
        return stats.system_memory_available

    def get_system_memory_percent(self, force_refresh: bool = False) -> float:
        """
        Get system memory usage percentage.

        Args:
            force_refresh: Force refresh of cached statistics

        Returns:
            float: System memory usage percentage

        """
        stats = self.get_memory_stats(force_refresh)
        return stats.system_memory_percent

    def is_memory_available(
        self, required_bytes: int, force_refresh: bool = False
    ) -> bool:
        """
        Check if required memory is available.

        Args:
            required_bytes: Required memory in bytes
            force_refresh: Force refresh of cached statistics

        Returns:
            bool: True if memory is available, False otherwise

        """
        stats = self.get_memory_stats(force_refresh)
        return stats.system_memory_available >= required_bytes

    def get_memory_summary(self, force_refresh: bool = False) -> dict[str, object]:
        """
        Get comprehensive memory summary.

        Args:
            force_refresh: Force refresh of cached statistics

        Returns:
            dict: Memory summary with all statistics

        """
        with self._lock:
            current_time = time.time()

            # Check if we should use cached data
            if (
                not force_refresh
                and self._cached_summary is not None
                and (current_time - self._last_update) < self.cache_duration
            ):
                # Return cached summary
                return self._cached_summary
            else:
                # Get fresh data
                stats = self.get_memory_stats(force_refresh)
                cache_age = current_time - stats.timestamp

                summary: dict[str, object] = {
                    "process_memory_bytes": stats.process_memory_bytes,
                    "process_memory_mb": stats.process_memory_mb,
                    "process_memory_percent": stats.process_memory_percent,
                    "system_memory_total": stats.system_memory_total,
                    "system_memory_total_gb": stats.system_memory_total_gb,
                    "system_memory_available": stats.system_memory_available,
                    "system_memory_available_gb": stats.system_memory_available_gb,
                    "system_memory_percent": stats.system_memory_percent,
                    "timestamp": stats.timestamp,
                    "cache_age": cache_age,
                }

                # Cache the summary
                self._cached_summary = summary
                return summary

    def clear_cache(self) -> None:
        """Clear cached memory statistics."""
        with self._lock:
            self._cached_stats = None
            self._cached_summary = None
            self._last_update = 0.0

    def __str__(self) -> str:
        """String representation of memory stats cache."""
        try:
            stats = self.get_memory_stats()
            return (
                f"MemoryStatsCache("
                f"process_memory={stats.process_memory_mb:.1f}MB, "
                f"system_memory={stats.system_memory_available_gb:.1f}GB "
                f"available)"
            )
        except Exception:
            return "MemoryStatsCache(unknown)"

    def __eq__(self, other: object) -> bool:
        """Compare memory stats cache based on configuration."""
        if not isinstance(other, MemoryStatsCache):
            return False
        return self.cache_duration == other.cache_duration

    def __hash__(self) -> int:
        """Hash memory stats cache based on configuration."""
        return hash(self.cache_duration)

    def __copy__(self) -> MemoryStatsCache:
        """Create a shallow copy of the memory stats cache."""
        return MemoryStatsCache(cache_duration=self.cache_duration)

    def __deepcopy__(self, memo: dict[str, object]) -> MemoryStatsCache:
        """Create a deep copy of the memory stats cache."""
        return MemoryStatsCache(cache_duration=self.cache_duration)


# Global memory stats cache instance
_global_memory_cache: MemoryStatsCache | None = None


def get_global_memory_cache() -> MemoryStatsCache:
    """
    Get the global memory stats cache instance.

    Returns:
        MemoryStatsCache: Global memory stats cache

    """
    global _global_memory_cache
    if _global_memory_cache is None:
        _global_memory_cache = MemoryStatsCache()
    return _global_memory_cache


def set_global_memory_cache(cache: MemoryStatsCache) -> None:
    """
    Set the global memory stats cache instance.

    Args:
        cache: Memory stats cache to use globally

    """
    global _global_memory_cache
    _global_memory_cache = cache
